####################################################################################################
# neuropythy/optimize/core.py
# Code for optimizing functions; made with the intention of optimizing functions on the cortical
# surface using gradient descent or a gradient-aware method.
# By Noah C. Benson

import os, gzip, types, six, abc, pimms
import numpy                 as np
import numpy.linalg          as npla
import scipy                 as sp
import scipy.sparse          as sps
import scipy.optimize        as spopt
import pyrsistent            as pyr
from   functools         import reduce
from   ..                import geometry as geo
from   ..                import mri      as mri
from   ..util            import (numel, rows, part, hstack, vstack, repmat, flatter, flattest,
                                 times, plus, minus, zdivide, zinv, power, ctimes, cpower, inner,
                                 cplus, sine, cosine, tangent, cosecant, secant, cotangent,
                                 divide, arctangent)
from   ..geometry        import (triangle_area)

# Helper Functions #################################################################################
def fapply(f, x, tz=False):
    '''
    fapply(f,x) yields the result of applying f either to x, if x is a normal value or array, or to
      x.data if x is a sparse matrix. Does not modify x (unless f modifiex x).

    The optional argument tz (default: False) may be set to True to specify that, if x is a sparse 
    matrix that contains at least 1 element that is a sparse-zero, then f(0) should replace all the
    sparse-zeros in x (unless f(0) == 0).
    '''
    if sps.issparse(x):
        y = x.copy()
        y.data = f(x.data)
        if tz and y.getnnz() < np.prod(y.shape):
            z = f(np.array(0))
            if z != 0:
                y = y.toarray()
                y[y == 0] = z
        return y
    else: return f(x)
def finto(x, ii, n, null=0):
    '''
    finto(x,ii,n) yields a vector u of length n such that u[ii] = x.

    Notes:
      * The ii index may be a tuple (as can be passed to numpy arrays' getitem method) in order to
        specify that the specific elements of a multidimensional output be set. In this case, the
        argument n should be a tuple of sizes (a single integer is taken to be a square/cube/etc).
      * x may be a sparse-array, but in it will be reified by this function.

    The following optional arguments are allowed:
      * null (defaut: 0) specifies the value that should appear in the elements of u that are not
        set.
    '''
    x  = x.toarray() if sps.issparse(x) else np.asarray(x)
    shx = x.shape
    if isinstance(ii, tuple):
        if not pimms.is_vector(n): n = tuple([n for _ in ii])
        if len(n) != len(ii): raise ValueError('%d-dim index but %d-dim output' % (len(ii),len(n)))
        sh = n + shx[1:]
    elif pimms.is_int(ii): sh = (n,) + shx
    else:                  sh = (n,) + shx[1:]
    u = np.zeros(sh, dtype=x.dtype) if null == 0 else np.full(sh, null, dtype=x.dtype)
    u[ii] = x
    return u

# Potential Functions ##############################################################################
@six.add_metaclass(abc.ABCMeta)
class PotentialFunction(object):
    '''
    The PotentialFunction class is intended as the base-class for all potential functions that can
    be minimized by neuropythy. PotentialFunction is effectively an abstract class that requires its
    subclasses to implement the method __call__(), which must take one argument: a numpy vector of
    parameters. The method must return a tuple of (z,dz) where z is  the potential value for the
    given paramters and dz is the Jacobian of the function at the parameters. Note that if the
    potential z returned is a scalar, then dz must be a vector of length len(params); if z is a
    vector, then dz must be a matrix of size (len(z) x len(params))
    '''
    # The value() function should return the potential alone
    @abc.abstractmethod
    def value(self, params):
        '''
        pf.value(params) yields the potential function value at the given parameters params. Values
          must always be vectors; if the function returns a scalar, it must return a 1-element
          vector instead.
        '''
        raise RuntimeError('The value() method was not overloaded for object %s' % self)
    @abc.abstractmethod
    def jacobian(self, params, into=None):
        '''
        pf.jacobian(params) yields the potential function jacobian matrix at the given parameters
          params. The jacobian must always be a (possibly sparse) matrix; if there is a scalar
          output value for the potential function, then Jacobian should be (1 x n); if there is a
          scalar input parameter, the Jacobian should be (n x 1).

        If the optional matrix into is provided then the returned jacobian may optionally be added
        directly into this matrix and returned.
        '''
        raise RuntimeError('The gradient() method was not overloaded for object %s' % self)
    # The __call__ function is how one generally calls a potential function
    def __call__(self, params):
        '''
        pf(params) yields the tuple (z, dz) where z is the potential value at the given parameters
          vector, params, and dz is the vector of the potential gradient.
        '''
        z  = self.value(params)
        dz = self.jacobian(params)
        if sps.issparse(dz): dz = dz.toarray()
        z  = np.squeeze(z)
        dz = np.squeeze(dz)
        return (z,dz)
    def fun(self):
        '''
        pf.fun() yields a value calculation function for the given potential function pf that is
          appropriate for passing to a minimizer.
        '''
        return lambda x: np.squeeze(self.value(x))
    def jac(self):
        '''
        pf.jac() yields a jacobian calculation function for the given potential function pf that is
          appropritate for passing to a minimizer.
        '''
        def _jacobian(x):
            dz = self.jacobian(x)
            if sps.issparse(dz): dz = dz.toarray()
            dz = np.asarray(dz)
            return np.squeeze(dz)
        return _jacobian
    def minimize(self, x0, **kwargs):
        '''
        pf.minimize(x0) minimizes the given potential function starting at the given point x0; any
          additional options are passed along to scipy.optimize.minimize.
        '''
        x0 = np.asarray(x0)
        kwargs = pimms.merge({'jac':self.jac(), 'method':'CG'}, kwargs)
        res = spopt.minimize(self.fun(), x0.flatten(), **kwargs)
        res.x = np.reshape(res.x, x0.shape)
        return res
    def argmin(self, x0, **kwargs):
        '''
        pf.argmin(x0) is equivalent to pf.minimize(x0).x.
        '''
        return self.minimize(x0, **kwargs).x
    def maximize(self, x0, **kwargs):
        '''
        pf.maximize(x0) is equivalent to (-pf).minimize(x0).
        '''
        return (-self).minimize(x0, **kwargs)
    def argmax(self, x0, **kwargs):
        '''
        pf.argmax(x0) is equivalent to pf.maximize(x0).x.
        '''
        return self.maximize(x0, **kwargs).x
    # Arithmetic Operators #########################################################################
    def __getitem__(self, ii):
        return part(self, ii)
    def __neg__(self):
        if is_const_potential(self): return const_potential(-self.c)
        else: return PotentialTimesConstant(self, -1)
    def __add__(self, x):
        x = to_potential(x)
        if is_const_potential(x):
            if   np.isclose(x.c, 0).all(): return self
            elif is_const_potential(self): return const_potential(self.c + x.c)
            else:                          return PotentialPlusConstant(self, x.c)
        elif is_const_potential(self):     return PotentialPlusConstant(x.c, self)
        else:                              return PotentialPlusPotential(self, x)
    def __radd__(self, x):
        return self.__add__(x)
    def __sub__(self, x):
        return self.__add__(-x)
    def __rsub__(self, x):
        return PotentialPlusConstant(-self, x)
    def __mul__(self, x):
        x = to_potential(x)
        if is_const_potential(x):
            if   np.isclose(x.c, 1).all(): return self
            elif np.isclose(x.c, 0).all(): return x
            elif is_const_potential(self): return const_potential(self.c * x.c)
            else:                          return PotentialTimesConstant(self, x.c)
        elif is_const_potential(self):     return PotentialTimesConstant(x, self.c)
        else:                              return PotentialTimesPotential(self, x)
    def __rmul__(self, x):
        return (self * x)
    def __div__(self, x):
        x = to_potential(x)
        if is_const_potential(x):
            if   np.isclose(x.c, 1).all(): return self
            elif is_const_potential(self): return const_potential(self.c / x.c)
            else:                          return PotentialTimesConstant(self, 1.0/x.c)
        else:                              return PotentialTimesPotential(self, 1/x)
    def __rdiv__(self, x):
        x = to_potential(x)
        if is_const_potential(self): return x / self.c
        else:                        return x * PotentialPowerConstant(self, -1)
    def __truediv__(self, x):
        return self.__div__(x)
    def __rtruediv__(self, x):
        return self.__rdiv__(x)
    def __pow__(self, x):
        x = to_potential(x)
        if is_const_potential(x):
            if   is_const_potential(self): return const_potential(power(self.c, x.c))
            else:                          return PotentialPowerConstant(self, x.c)
        elif is_const_potential(self):     return ConstantPowerPotential(self, x)
        else:                              return PotentialPowerPotential(self, x)
    def __rpow__(self, x):
        return to_potential(x).__pow__(self)
    def compose(self, f):
        return compose(self, f)
def is_potential(f):
    '''
    is_potential(f) yields True if f is a potential function and False otherwise.
    '''
    return isinstance(f, PotentialFunction)
def safe_into(into, term):
    if into is None: return term
    into0 = into
    into += term
    if into is into0: return into
    else: return term
@pimms.immutable
class PotentialIdentity(PotentialFunction):
    '''
    PotentialIdentity is a potential function that represents the arguments given to it as outputs.
    '''
    def __init__(self): pass
    def value(self, params): return np.asarray(params)
    def jacobian(self, params, into=None):
        return safe_into(into, sps.eye(numel(params)))
identity = PotentialIdentity()
def is_identity_potential(f):
    '''
    is_identity_potential(f) yields True if f is a potential function that merely yields its
      parameters (f(x) = x); otherwise yields False.
    '''
    return isinstance(f, PotentialIdentity)
@pimms.immutable
class PotentialLambda(PotentialFunction):
    '''
    PotentialLambda is a PotentialFunction type that takes as initialization arguments two functions
    g and h which must return the value and the jacobian, respectively.
    '''
    def __init__(self, g, h):
        self.valfn = g
        self.jacfn = h
    @pimms.param
    def valfn(v): return v
    @pimms.param
    def jacfn(j): return j
    def value(self, x): return self.valfn(x)
    def jacobian(self, x, into=None):
        try:              return self.jacfn(x, into)
        except Exception: return self.jacfn(x)
@pimms.immutable
class PotentialConstant(PotentialFunction):
    def __init__(self, c):
        self.c = c
    @pimms.param
    def c(c0):
        c0 = c0.toarray() if sps.issparse(c0) else np.array(c0)
        if len(c0.shape) > 1: c0 = c0.flatten()
        c0.setflags(write=False)
        return c0
    def value(self, params):
        return self.c
    def jacobian(self, params, into=None):
        c = self.c
        d = 1 if len(c.shape) == 0 else c.shape[0]
        return sps.csr_matrix(([], [[],[]]), shape=(d, len(params))) if into is None else into
def is_const_potential(f):
    '''
    is_const_potential(f) yields True if f is a constant potential function and False otherwise.
    '''
    return isinstance(f, PotentialConstant)
def const_potential(f):
    '''
    const_potential(f) yields f if f is a constant potential function; if f is a constant, yields
      a potential function that always yields f; otherwise raises an error.
    '''
    if is_const_potential(f): return f
    elif pimms.is_array(f, 'number'): return PotentialConstant(f)
    else: raise ValueError('Could not convert given value to potential constant: %s' % f)
def const(f):
    '''
    const(f) yields f.c if f is a constant potential function; if f is a constant, it yields f or
      the equivalent numpy array object; if f is a potential function that is not const, or is not
      a valid potential function constant, yields None.
    '''
    if is_const_potential(f): return f.c
    elif pimms.is_array(f, 'number'):
        if sps.issparse(f): return f
        else: return np.asarray(f)
    else: return None
def to_potential(f):
    '''
    to_potential(f) yields f if f is a potential function; if f is not, but f can be converted to
      a potential function, that conversion is performed then the result is yielded.
    to_potential(Ellipsis) yields a potential function whose output is simply its input (i.e., the
      identity function).
    to_potential(None) is equivalent to to_potential(0).

    The following can be converted into potential functions:
      * Anything for which pimms.is_array(x, 'number') yields True (i.e., arrays of constants).
      * Any tuple (g, h) where g(x) yields a potential value and h(x) yields a jacobian matrix for
        the parameter vector x.
    '''
    if   is_potential(f): return f
    elif f is Ellipsis:   return identity
    elif pimms.is_array(f, 'number'): return const_potential(f)
    elif isinstance(f, tuple) and len(f) == 2: return PotentialLambda(f[0], f[1])
    else: raise ValueError('Could not convert object of type %s to potential function' % type(f))
@pimms.immutable
class PotentialComposition(PotentialFunction):
    def __init__(self, g, h):
        self.g = g
        self.h = h
    @pimms.param
    def g(g0): return g0
    @pimms.param
    def h(h0): return h0
    def value(self, params):
        return self.g.value(self.h.value(params))
    def jacobian(self, params, into=None):
        zh  = self.h.value(params)
        dzh = self.h.jacobian(params)
        zg  = self.g.value(zh)
        dzg = self.g.jacobian(zh)
        return safe_into(into, inner(dzg, dzh))
def compose(*args):
    '''
    compose(g, h...) yields a potential function f that is the result of composing together all the
      arguments g, h, etc. after calling to_potential() on each. The result is defined such that
      f(x) is equivalent to g(h(...(x))).
    '''
    return reduce(lambda h,g: PotentialComposition(g,h), reversed(list(map(to_potential, args))))
@pimms.immutable
class PotentialPart(PotentialFunction):
    def __init__(self, ii, input_len=None):
        self.output_indices = ii
        self.input_len = input_len
    @pimms.param
    def output_indices(ii):
        ii = flattest(ii)
        if (np.issubdtype(ii.dtype, np.dtype('bool').type)): ii = np.where(ii)[0]
        return pimms.imm_array(ii)
    @pimms.param
    def input_len(m):
        if m is None: return m
        assert(pimms.is_int(m) and m > 0)
        return int(m)
    @pimms.value
    def jacobian_matrix(output_indices, input_len):
        m = (np.max(output_indices) + 1) if input_len is None else input_len
        n = len(output_indices)
        return sps.csr_matrix((np.ones(n), (np.arange(n), output_indices)), shape=(n,m))
    def value(self, params):
        ii = self.output_indices
        return flattest(params)[ii]
    def jacobian(self, params, into=None):
        params = flattest(params)
        jm = self.jacobian_matrix
        if jm.shape[1] != len(params):
            jm = jm.copy()
            jm.resize((jm.shape[0], len(params)))
        return safe_into(into, jm)
def part(f, ii=None, input_len=None):
    '''
    part(u, ii) for constant or constant potential u yields a constant-potential form of u[ii].
    part(f, ii) for potential function f yields a potential function g(x) that is equivalent to
      f(x)[ii].
    part(ii) is equivalent to part(identity, ii); i.e., pat of the input parameters to the function.
    '''
    if ii is None: return PotentialPart(f, input_len=input_len)
    f = to_potential(f)
    if is_const_potential(f): return PotentialConstant(f.c[ii])
    else:                     return compose(PotentialPart(ii, input_len=input_len), f)
@pimms.immutable
class PotentialPlusPotential(PotentialFunction):
    def __init__(self, g, h):
        self.g = g
        self.h = h
    @pimms.param
    def g(g0): return g0
    @pimms.param
    def h(h0): return h0
    def value(self, params):
        return self.g.value(params) + self.h.value(params)
    def jacobian(self, params, into=None):
        dg = self.g.jacobian(params, into=into)
        dh = self.h.jacobian(params, into=dg)
        if dh is dg: return dh
        else:        return dh + dg
@pimms.immutable
class PotentialPlusConstant(PotentialFunction):
    def __init__(self, f, c):
        self.f = f
        self.c = c
    @pimms.param
    def f(f0): return f0
    @pimms.param
    def c(c0): return c0
    def value(self,params):
        return self.f.value(params) + self.c
    def jacobian(self, params, into=None):
        return self.f.jacobian(params, into=into)
@pimms.immutable
class PotentialTimesPotential(PotentialFunction):
    def __init__(self, g, h):
        self.g = g
        self.h = h
    @pimms.param
    def g(g0): return g0
    @pimms.param
    def h(h0): return h0
    def value(self, params):
        g = self.g.value(params)
        h = self.h.value(params)
        return g * h
    def jacobian(self, params, into=None):
        g  = self.g.value(params)
        dg = self.g.jacobian(params)
        h  = self.h.value(params)
        dh = self.h.jacobian(params)
        return safe_into(into, cplus(times(dg, h), times(dh, g)))
    def __call__(self, params):
        g  = self.g.value(params)
        dg = self.g.jacobian(params)
        h  = self.h.value(params)
        dh = self.h.jacobian(params)
        return (g*h, cplus(times(dg, h), times(dh, g)))
@pimms.immutable
class PotentialTimesConstant(PotentialFunction):
    def __init__(self, f, c):
        self.f = f
        self.c = c
    @pimms.param
    def f(f0): return f0
    @pimms.param
    def c(c0): return c0
    def value(self, params):
        z = self.f.value(params)
        return z * self.c
    def jacobian(self, params, into=None):
        dz = self.f.jacobian(params)
        return safe_into(into, times(dz, self.c))
@pimms.immutable
class PotentialPowerConstant(PotentialFunction):
    def __init__(self, f, c):
        self.f = f
        self.c = c
    @pimms.param
    def f(f0): return f0
    @pimms.param
    def c(c0): return c0
    def value(self, params):
        z = self.f.value(params)
        return z**self.c
    def jacobian(self, params, into=None):
        z  = self.f.value(params)
        dz = self.f.jacobian(params)
        c  = self.c
        cc = self.c - 1
        if cc <= 0:
            cc = -cc
            z = zinv(z)
        return safe_into(into, times(dz, c * z**cc))
    def __call__(self, params):
        z  = self.f.value(params)
        dz = self.f.jacobian(params)
        c  = self.c
        dz = times(dz, c * z**(c-1))
        return (z, dz)
@pimms.immutable
class ConstantPowerPotential(PotentialFunction):
    def __init__(self, c, f):
        self.f = f
        self.c = c
    @pimms.param
    def f(f0): return f0
    @pimms.param
    def c(c0):
        if np.isclose(c0, 0).any(): raise ValueError('in c**f(x) c cannot be zero')
        return c0
    @pimms.value
    def log_c(c): return np.log(c)
    def value(self, params):
        z = self.f.value(params)
        return self.c**z
    def jacobian(self, params, into=None):
        ctoz = self.value(params)
        dz = self.f.jacobian(params)
        return safe_into(into, times(dz, self.log_c * ctoz))
def exp(x):
    x = to_potential(x)
    if is_const_potential(x): return PotentialConstant(np.exp(x.c))
    else:                     return ConstantPowerPotential(np.e, x)
def exp2(x):
    x = to_potential(x)
    if is_const_potential(x): return PotentialConstant(np.exp2(x.c))
    else:                     return ConstantPowerPotential(2.0, x)
@pimms.immutable
class PotentialPowerPotential(PotentialFunction):
    def __init__(self, g, h):
        self.g = g
        self.h = h
    @pimms.param
    def g(g0): return g0
    @pimms.param
    def h(h0): return h0
    def value(self, params):
        zg = self.g.value(params)
        zh = self.h.value(params)
        return zg ** zh
    def jacobian(self, params, into=None):
        zg  = self.g.value(params)
        dzg = self.g.jacobian(params)
        zh  = self.h.value(params)
        dzh = self.h.jacobian(params)
        z   = zg ** zh
        return safe_into(into, times(plus(times(dzg, zh, inv(zg)), times(dzh, np.log(zg))), z))
def power(x,y):
    x = to_potential(x)
    y = to_potential(y)
    xc = is_const_potential(x)
    yc = is_const_potential(y)
    if xc and yc: return PotentialPowerPotential(x,   y)
    elif xc:      return ConstantPowerPotential( x.c, y)
    elif yc:      return PotentialPowerConstant( x,   y.c)
    else:         return PotentialConstant(power(x.c, y.c))
def sqrt(x): return power(x, 0.5)
def cbrt(x): return power(x, 1.0/3.0)
@pimms.immutable
class PotentialLog(PotentialFunction):
    def __init__(self, f, base=None):
        self.f = f
        self.base = base
    @pimms.param
    def f(f0): return to_potential(f0)
    @pimms.param
    def base(b): return None if b is None else to_potential(b)
    def value(self, params):
        z = self.f.value(params)
        if self.base is None: return np.log(z)
        b = self.base.value(params)
        return np.log(z)/np.log(b)
    def jacobian(self, params, into=None):
        z  = self.f.value(params)
        dz = self.f.jacobian(params)
        if self.base is None:
            dz = divide(dz, z)
        else:
            b = self.base.value(params)
            db = self.base.jacobian(params)
            logb = np.log(b)
            dz = dz / logb - times(np.log(z), db) / (b * logb * logb)
        return safe_into(into, dz)
def log(x, base=None):
    x = to_potential(x)
    xc = is_const_potential(x)
    if base is None:
        if xc: return PotentialConstant(np.log(x.c))
        else:  return PotentialLog(x)
    base = to_potential(base)
    bc = is_const_potential(base)
    if xc and bc: return PotentialConstant(np.log(x.c, bc.c))
    else:         return PotentialLog(x, base)
def log2(x):  return log(x, 2)
def log10(x): return log(x, 10)
@pimms.immutable
class PotentialSum(PotentialFunction):
    def __init__(self, f, weights=None):
        self.f = f
        self.weights = None
    @pimms.param
    def f(f0): return to_potential(f0)
    @pimms.param
    def weights(w): return None if w is None else pimms.imm_array(w)
    def value(self, params):
        z = self.f.value(params)
        w = self.weights
        if w is None: return np.sum(z)
        else: return np.dot(z, w)
    def jacobian(self, params, into=None):
        dz = self.f.jacobian(params)
        w = self.weights
        if w is None: q = dz.sum(axis=0)
        else:         q = times(dz, w).sum(axis=0)
        return safe_into(into, q)
def sum(x, weights=None):
    '''
    sum(x) yields either a potential-sum object if x is a potential function or the sum of x if x
      is not. If x is not a potential-field then it must be a vector.
    sum(x, weights=w) uses the given weights to produce a weighted sum.
    '''
    x = to_potential(x)
    if is_const_potential(x): return PotentialConstant(np.sum(x.c))
    else:                     return PotentialSum(x, weights=weights)
@pimms.immutable
class DotPotential(PotentialFunction):
    '''
    DotPotential is a potential function that represents the dot product of two functions.
    '''
    def __init__(self, g, h, g_shape=None, h_shape=None):
        self.g = g
        self.h = h
        self.g_shape = g_shape
        self.h_shape = h_shape
    @pimms.param
    def g(g0): return to_potential(g0)
    @pimms.param
    def h(h0): return to_potential(h0)
    @pimms.param
    def g_shape(gs):
        if gs is None: return None
        gs = tuple(gs)
        if   len(gs) < 2:  return None
        #elif len(gs) == 2: return gs
        else: raise ValueError('dot supports only scalars, vectors, and (soon) matrices')
    @pimms.param
    def h_shape(hs):
        if hs is None: return None
        hs = tuple(hs)
        if   len(hs) < 2:  return None
        #elif len(hs) == 2: return hs
        else: raise ValueError('dot supports only scalars, vectors, and (soon) matrices')
    def value(self, params):
        g = self.g.value(params)
        h = self.h.value(params)
        g = np.reshape(g, self.g_shape) if self.g_shape else flattest(g)
        h = np.reshape(h, self.h_shape) if self.h_shape else flattest(h)
        return flattest(inner(g, h))
    def jacobian(self, params, into=None):
        g = self.g.value(params)
        h = self.h.value(params)
        g = np.reshape(g, self.g_shape) if self.g_shape else flattest(g)
        h = np.reshape(h, self.h_shape) if self.h_shape else flattest(h)
        dg = self.g.jacobian(params)
        dh = self.h.jacobian(params)
        gvec = self.g_shape is None
        hvec = self.h_shape is None
        if gvec == hvec:
            if gvec: return np.sum(g*dh + h*dg)
        # one or both are matrices
        raise NotImplementedError('matrix x matrix dot products not yet supported')
def dot(a, b, ashape=None, bshape=None):
    '''
    dot(a,b) yields a potential function that represents the dot product of a and b.

    Currently only vector and scalars are allowed.
    '''
    a = to_potential(a)
    b = to_potential(b)
    if is_const_potential(a) and is_const_potential(b): return PotentialConstant(np.dot(a.c, b.c))
    else: return DotPotential(a, b, g_shape=ashape, h_shape=bshape)
@pimms.immutable
class CosPotential(PotentialFunction):
    '''
    CosPotential is a potential function that represents cos(x).
    '''
    def __init__(self): pass
    def value(self, x): return cosine(x)
    def jacobian(self, x, into=None):
        x = flattest(x)
        z = sps.diags(-sine(x))
        return safe_into(into, z)
@pimms.immutable
class SinPotential(PotentialFunction):
    '''
    SinPotential is a potential function that represents sin(x).
    '''
    def __init__(self): pass
    def value(self, x): return sine(x)
    def jacobian(self, x, into=None):
        x = flattest(x)
        z = sps.diags(cosine(x))
        return safe_into(into, z)
@pimms.immutable
class TanPotential(PotentialFunction):
    '''
    TanPotential is a potential function that represents tan(x).
    '''
    def __init__(self): pass
    def value(self, x): return tangent(x)
    def jacobian(self, x, into=None):
        x = flattest(x)
        z = sps.diags(secant(x)**2)
        return safe_into(into, z)
@pimms.immutable
class SecPotential(PotentialFunction):
    '''
    SecPotential is a potential function that represents sec(x).
    '''
    def __init__(self): pass
    def value(self, x): return secant(x)
    def jacobian(self, x, into=None):
        x = flattest(x)
        z = sps.diags(secant(x)*tangent(x))
        return safe_into(into, z)
@pimms.immutable
class CscPotential(PotentialFunction):
    '''
    CscPotential is a potential function that represents csc(x).
    '''
    def __init__(self): pass
    def value(self, x): return cosecant(x)
    def jacobian(self, x, into=None):
        x = flattest(x)
        z = sps.diags(-cosecant(x)*cotangent(x))
        return safe_into(into, z)
@pimms.immutable
class CotPotential(PotentialFunction):
    '''
    CotPotential is a potential function that represents cot(x).
    '''
    def __init__(self): pass
    def value(self, x): return cotangent(x)
    def jacobian(self, x, into=None):
        x = flattest(x)
        return safe_into(into, -cosecant(x)**2)
def cos(x):
    x = to_potential(x)
    if is_const_potential(x): return PotentialConstant(cosine(x.c))
    elif x is identity:       return CosPotential()
    else:                     return compose(CosPotential(), x)
def sin(x):
    x = to_potential(x)
    if is_const_potential(x): return PotentialConstant(sine(x.c))
    elif x is identity:       return SinPotential()
    else:                     return compose(SinPotential(), x)
def tan(x):
    x = to_potential(x)
    if is_const_potential(x): return PotentialConstant(tangent(x.c))
    elif x is identity:       return TanPotential()
    else:                     return compose(TanPotential(), x)
def sec(x):
    x = to_potential(x)
    if is_const_potential(x): return PotentialConstant(secant(x.c))
    elif x is identity:       return SecPotential()
    else:                     return compose(SecPotential(), x)
def csc(x):
    x = to_potential(x)
    if is_const_potential(x): return PotentialConstant(cosecant(x.c))
    elif x is identity:       return CscPotential()
    else:                     return compose(CscPotential(), x)
def cot(x):
    x = to_potential(x)
    if is_const_potential(x): return PotentialConstant(cotangent(x.c))
    elif x is identity:       return CotPotential()
    else:                     return compose(CotPotential(), x)
@pimms.immutable
class ArcSinPotential(PotentialFunction):
    '''
    ArcSinPotential is a potential function that represents asin(x).
    '''
    def __init__(self): pass
    def value(self, x): return arcsine(x)
    def jacobian(self, x, into=None):
        x = flattest(x)[None]
        z = 1.0 / np.sqrt(1.0 - x**2)
        z = sps.diags(z)
        return safe_into(into, z)
@pimms.immutable
class ArcCosPotential(PotentialFunction):
    '''
    ArcCosPotential is a potential function that represents acos(x).
    '''
    def __init__(self): pass
    def value(self, x): return arccosine(x)
    def jacobian(self, x, into=None):
        x = flattest(x)[None]
        z = -1.0 / np.sqrt(1.0 - x**2)
        z = sps.diags(z)
        return safe_into(into, z)
@pimms.immutable
class ArcTanPotential(PotentialFunction):
    '''
    ArcTanPotential is a potential function that represents atan(x).
    '''
    def __init__(self): pass
    def value(self, x): return arctangent(x)
    def jacobian(self, x, into=None):
        x = flattest(x)[None]
        z = 1.0 / (1.0 + x**2)
        z = sps.diags(z)
        return safe_into(into, z)
@pimms.immutable
class ArcTan2Potential(PotentialFunction):
    '''
    ArcTan2Potential is a potential function that represents atan2(y,x).
    '''
    def __init__(self, y, x):
        self.y = y
        self.x = x
    @pimms.param
    def y(y0): return to_potential(y0)
    @pimms.param
    def x(x0): return to_potential(x0)
    def value(self, params):
        y = self.y.value(params)
        x = self.x.value(params)
        return arctangent(y, x)
    def jacobian(self, params, into=None):
        y  = self.y.value(params)
        x  = self.x.value(params)
        dy = self.y.jacobian(params)
        dx = self.x.jacobian(params)
        if   dy.shape[0] == 1 and dx.shape[0] > 1: dy = repmat(dy, dx.shape[0], 1)
        elif dx.shape[0] == 1 and dy.shape[0] > 1: dx = repmat(dx, dy.shape[0], 1)
        dz = zdivide(times(dy, x) - times(dx, y), x**2 + y**2)
        return safe_into(into, dz)
def asin(x):
    x = to_potential(x)
    if is_const_potential(x): return PotentialConstant(arcsine(x.c))
    elif x is identity:       return ArcSinPotential()
    else:                     return compose(ArcSinPotential(), x)
def acos(x):
    x = to_potential(x)
    if is_const_potential(x): return PotentialConstant(arccosine(x.c))
    elif x is identity:       return ArcCosPotential()
    else:                     return compose(ArcCosPotential(), x)
def atan(x):
    x = to_potential(x)
    if is_const_potential(x): return PotentialConstant(arctangent(x.c))
    elif x is identity:       return ArcTanPotential()
    else:                     return compose(ArcTanPotential(), x)
def atan2(y,x): return ArcTan2Potential(y, x)
def row_norms(ii, f=Ellipsis, squared=False):
    '''
    row_norms(ii) yields a potential function h(x) that calculates the vector norms of the rows of
      the matrix formed by [x[i] for i in ii] (ii is a matrix of parameter indices).
    row_norms(ii, f) yield a potential function h(x) equivalent to compose(row_norms(ii), f).
    '''
    try:
        (n,m) = ii
        # matrix shape given
        ii = np.reshape(np.arange(n*m), (n,m))
    except Exception: ii = np.asarray(ii)
    f = to_potential(f)
    if is_const_potential(f):
        q = flattest(f.c)
        q = np.sum([q[i]**2 for i in ii.T], axis=0)
        return PotentialConstant(q if squared else np.sqrt(q))
    F = reduce(lambda a,b: a + b, [part(Ellipsis, col)**2 for col in ii.T])
    F = compose(F, f)
    if not squared: F = sqrt(F)
    return F
def col_norms(ii, f=Ellipsis, squared=False):
    '''
    col_norms(ii) yields a potential function h(x) that calculates the vector norms of the columns
      of the matrix formed by [x[i] for i in ii] (ii is a matrix of parameter indices).
    col_norms(ii, f) yield a potential function h(x) equivalent to compose(col_norms(ii), f).
    '''
    try:
        (n,m) = ii
        # matrix shape given
        ii = np.reshape(np.arange(n*m), (n,m))
    except Exception: ii = np.asarray(ii)
    f = to_potential(f)
    if is_const_potential(f):
        q = flattest(f.c)
        q = np.sum([q[i]**2 for i in ii], axis=0)
        return PotentialConstant(q if squared else np.sqrt(q))
    F = reduce(lambda a,b: a + b, [part(Ellipsis, col)**2 for col in ii])
    F = compose(F, f)
    if not squared: F = sqrt(F)
    return F
def distances(a, b, shape, squared=False, axis=1):
    '''
    distances(a, b, (n,d)) yields a potential function whose output is equivalent to the row-norms
      of reshape(a(x), (n,d)) - reshape(b(x), (n,d)).
    
    The shape argument (n,m) may alternately be a matrix of parameter indices, as can be passed to
    row_norms and col_norms.

    The following optional arguments are accepted:
      * squared (default: False) specifies whether the output should be the square distance or the
        distance.
      * axis (default: 1) specifies whether the rows (axis = 1) or columns (axis = 0) are treated
        as the vectors between which the distances should be calculated.
    '''
    a = to_potential(a)
    b = to_potential(b)
    if axis == 1: return row_norms(shape, a - b, squared=squared)
    else:         return col_norms(shape, a - b, squared=squared)
@pimms.immutable
class PotentialPiecewise(PotentialFunction):
    def __init__(self, dflt, *args):
        self.default = dflt
        self.pieces = args
    @pimms.param
    def default(d): return to_potential(d)
    @pimms.param
    def pieces(ps):
        r = []
        for p in ps:
            try:              ((mn, mx), f) = p
            except Exception: (mn,  mx,  f) = p
            if mx < mn: raise ValueError('given piece has mn > mx: %s' % p)
            f = to_potential(f)
            r.append(((mn,mx),f))
        r = sorted(r, key=lambda x:x[0])
        for (((lmn,lmx),_),((umn,umx),_)) in zip(r[:-1],r[1:]):
            if lmx > umn: raise ValueError('pieces contain overlapping ranges')
        return tuple(r)
    @pimms.value
    def pieces_with_default(pieces, default):
        ps = list(pieces)
        ps.append(((-np.inf,np.inf), default))
        return tuple(ps)
    def value(self, params):
        params = flattest(params)
        n = len(params)
        ii = np.arange(n)
        res = np.zeros(n)
        for ((mn,mx), f) in self.pieces_with_default:
            if len(ii) == 0: break
            k = np.where((params >= mn) & (params <= mx))[0]
            if len(k) == 0: continue
            kk = ii[k]
            res[kk] = f.value(params[k])
            ii = np.delete(ii, k)
            params = np.delete(params, k)
        return res
    def jacobian(self, params, into=None):
        params = flattest(params)
        n = len(params)
        ii = np.arange(n)
        (rs,cs,zs) = ([],[],[])
        for ((mn,mx), f) in self.pieces_with_default:
            if len(ii) == 0: break
            k = np.where((params >= mn) & (params <= mx))[0]
            if len(k) == 0: continue
            kk = ii[k]
            j = f.jacobian(params[k])
            if j.shape[0] == 1 and j.shape[1] > 1: j = repmat(j, j.shape[1], 1)
            (rj,cj,vj) = sps.find(j)
            rs.append(kk[rj])
            cs.append(kk[cj])
            zs.append(vj)
            ii = np.delete(ii, k)
            params = np.delete(params, k)
        (rs,cs,zs) = [np.concatenate(us) if len(us) > 0 else [] for us in (rs,cs,zs)]
        dz = sps.csr_matrix((zs, (rs,cs)), shape=(n,n))
        return safe_into(into, dz)
def piecewise(dflt, *spec):
    '''
    piecewise(g, ((mn1, mx1), f1), ((mn2, mx2), f2), ...) yields a potential function f(x) that, for
      each value x[i] in x, calculate y[i] = f1(x[i]) if mn1 <= x[i] <= mx1 else f2(x[i]) if mn2 <=
      x[i] <= mx2 else ... else g(x[i]).

    The ((mn,mx), f) may alternately be specified (mn,mx,f).
    '''
    return PotentialPiecewise(dflt, *spec)
def cos_well(f=Ellipsis, width=np.pi/2, offset=0, scale=1):
    '''
    cos_well() yields a potential function g(x) that calculates 0.5*(1 - cos(x)) for -pi/2 <= x
      <= pi/2 and is 1 outside of that range.
    
    The full formulat of the cosine well is, including optional arguments:
      scale / 2 * (1 - cos((x - offset) / (width/pi)))

    The following optional arguments may be given:
      * width (default: pi) specifies that the frequency of the cos-curve should be pi/width; the
        width is the distance between the points on the cos-curve with the value of 1.
      * offset (default: 0) specifies the offset of the minimum value of the coine curve on the
        x-axis.
      * scale (default: 1) specifies the height of the cosine well.
    '''
    f = to_potential(f)
    freq = np.pi/width*2
    (xmn,xmx) = (offset - width/2, offset + width/2)
    F = piecewise(scale, ((xmn,xmx), scale/2 * (1 - cos(freq * (identity - offset)))))
    if   is_const_potential(f):    return const_potential(F.value(f.c))
    elif is_identity_potential(f): return F
    else:                          return compose(F, f)
def cos_edge(f=Ellipsis, width=np.pi, offset=0, scale=1):
    '''
    cos_edge() yields a potential function g(x) that calculates 0 for x < pi/2, 1 for x > pi/2, and
      0.5*(1 + cos(pi/2*(1 - x))) for x between -pi/2 and pi/2.
    
    The full formulat of the cosine well is, including optional arguments:
      scale/2 * (1 + cos(pi*(0.5 - (x - offset)/width)

    The following optional arguments may be given:
      * width (default: pi) specifies that the frequency of the cos-curve should be pi/width; the
        width is the distance between the points on the cos-curve with the value of 1.
      * offset (default: 0) specifies the offset of the minimum value of the coine curve on the
        x-axis.
      * scale (default: 1) specifies the height of the cosine well.
    '''
    f = to_potential(f)
    freq = np.pi/2
    (xmn,xmx) = (offset - width/2, offset + width/2)
    F = piecewise(scale,
                  ((-np.inf, xmn), 0),
                  ((xmn,xmx), scale/2 * (1 + cos(np.pi*(0.5 - (identity - offset)/width)))))
    if   is_const_potential(f):    return const_potential(F.value(f.c))
    elif is_identity_potential(f): return F
    else:                          return compose(F, f)
def gaussian(f=Ellipsis, mu=0, sigma=1, scale=1, invert=False, normalize=False):
    '''
    gaussian() yields a potential function f(x) that calculates a Gaussian function over x; the
      formula used is given below.
    gaussian(g) yields a function h(x) such that, if f(x) is yielded by gaussian(), h(x) = f(g(x)).

    The formula employed by the Gaussian function is as follows, with mu, sigma, and scale all being
    parameters that one can provide via optional arguments:
      scale * exp(0.5 * ((x - mu) / sigma)**2)
    
    The following optional arguments may be given:
      * mu (default: 0) specifies the mean of the Gaussian.
      * sigma (default: 1) specifies the standard deviation (sigma) parameger of the Gaussian.
      * scale (default: 1) specifies the scale to use.
      * invert (default: False) specifies whether the Gaussian should be inverted. If inverted, then
        the formula, scale * exp(...), is replaced with scale * (1 - exp(...)).
      * normalize (default: False) specifies whether the result should be multiplied by the inverse
        of the area under the uninverted and unscaled curve; i.e., if normalize is True, the entire
        result is multiplied by 1/sqrt(2*pi*sigma**2).
    '''
    f = to_potential(f)
    F = exp(-0.5 * ((f - mu) / sigma)**2)
    if invert: F = 1 - F
    F = F * scale
    if normalize: F = F / (np.sqrt(2.0*np.pi) * sigma)
    return F
@pimms.immutable
class ErfPotential(PotentialFunction):
    '''
    ErfPotential is a potential function that represents the error function.
    '''
    coef = 2.0 / np.sqrt(np.pi)
    def __init__(self): pass
    def value(self, x): return np.erf(flattest(x))
    def jacobian(self, x, into=None):
        x = flattest(x)
        z = ErfPotential.coef * np.exp(-x**2)
        z = sps.diags(z)
        return safe_into(into, z)
def erf(f=Ellipsis):
    '''
    erf(x) yields a potential function that calculates the error function over the input x. If x is
      a constant, yields a constant potential function.
    erf() is equivalent to erf(...), which is just the error function, calculated over its inputs.
    '''
    f = to_potential(f)
    if is_const_potential(f): return const_potential(np.erf(f.c))
    elif is_identity_potential(f): return ErfPotential()
    else: return compose(ErfPotential(), f)
def sigmoid(f=Ellipsis, mu=0, sigma=1, scale=1, invert=False, normalize=False):
    '''
    sigmoid() yields a potential function that is equivalent to the integral of gaussian(), i.e.,
      the error function, but scaled to match gaussian().
    sigmoid(f) is equivalent to compose(sigmoid(), f).

    All options that are accepted by the gaussian() function are accepted by sigmoid() with the same
    default values and are handled in an equivalent manner with the exception of the invert option;
    when a sigmoid is inverted, the function approaches its maximum value at -inf and approaches 0
    at inf.

    Note that because sigmoid() explicitly matches gaussian(), the base formula used is as follows:
      f(x) = scale * sigma * sqrt(pi/2) * erf((x - mu) / (sqrt(2) * sigma))
      k*sig*Sqrt[Pi/2] Erf[(x - mu)/sig/Sqrt[2]]
    '''
    f = to_potential(f)
    F = erf((f - mu) / (sigma * np.sqrt(2.0)))
    if invert: F = 1 - F
    F = np.sqrt(np.pi / 2) * scale * F
    if normalize: F = F / (np.sqrt(2.0*np.pi) * sigma)
    return F
@pimms.immutable
class AbsPotential(PotentialFunction):
    '''
    AbsPotential is a potential function that represents the absolute value function.
    '''
    def __init__(self): pass
    def value(self, x): return np.abs(flattest(x))
    def jacobian(self, x, into=None):
        x = flattest(x)
        z = np.sign(x)
        z = sps.diags(z)
        return safe_into(into, z)
def abs(f=Ellipsis):
    '''
    abs() yields a potential function equivalent to the absolute value of the input.
    abs(f) yields the absolute value of the potential function f.

    Note that abs has a derivative of 0 at 0; this is not mathematically correct, but it is useful
    for the purposes of numerical methods. If you want traditional behavior, it is suggested that
    one instead employ sqrt(f**2).
    '''
    f = to_potential(f)
    if is_const_potential(f): return const_potential(np.abs(f.c))
    elif is_identity_potential(f): return AbsPotential()
    else: return compose(AbsPotential(), f)
@pimms.immutable
class SignPotential(PotentialFunction):
    '''
    SignPotential is a potential function that represents the sign function.
    '''
    def __init__(self): pass
    def value(self, x): return np.sign(flattest(x))
    def jacobian(self, x, into=None):
        n = len(flattest(x))
        return sps.csr_matrix(([], [[],[]]), shape=(n,n)) if into is None else into
def sign(f=Ellipsis):
    '''
    sign() yields a potential function equivalent to the sign of the input.
    sign(f) yields the sign of the potential function f.

    Note that sign has a derivative of 0 at all points; this is not mathematically correct, but it is
    useful for the purposes of numerical methods. If you want traditional behavior, it is suggested
    that one instead employ f/sqrt(f**2).
    '''
    f = to_potential(f)
    if is_const_potential(f): return const_potential(np.sign(f.c))
    elif is_identity_potential(f): return SignPotential()
    else: return compose(SignPotential(), f)
    
@pimms.immutable
class TriangleSignedArea2DPotential(PotentialFunction):
    '''
    TriangleSignedArea2DPotential(n) yields a potential function that tracks the signed area of
    the given face count n embedded in 2 dimensions.
    The signed area is positive if the triangle is counter-clockwise and negative if the triangle is
    clockwise.
    '''
    def value(self, p):
        # transpose to be 3 x 2 x n
        p = np.transpose(np.reshape(p, (-1, 3, 2)), (1,2,0))
        # First, get the two legs...
        (dx_ab, dy_ab) = p[1] - p[0]
        (dx_ac, dy_ac) = p[2] - p[0]
        (dx_bc, dx_bc) = p[2] - p[1]
        # now, the area is half the z-value of the cross-product...
        sarea = 0.5 * (dx_ab*dy_ac - dx_ac*dy_ab)
        return sarea
    def jacobian(self, p, into=None):
        p = np.transpose(np.reshape(p, (-1, 3, 2)), (1,2,0))
        (dx_ab, dy_ab) = p[1] - p[0]
        (dx_ac, dy_ac) = p[2] - p[0]
        (dx_bc, dy_bc) = p[2] - p[1]
        z = 0.5 * np.transpose([[-dy_bc,dx_bc], [dy_ac,-dx_ac], [-dy_ab,dx_ab]], (2,0,1))
        m = numel(p)
        n = p.shape[2]
        ii = (np.arange(n) * np.ones([6, n])).T.flatten()
        z = sps.csr_matrix((z.flatten(), (ii, np.arange(len(ii)))), shape=(n, m))
        return safe_into(into, z)
def signed_face_areas(faces, axis=1):
    '''
    signed_face_areas(faces) yields a potential function f(x) that calculates the signed area of
      each face represented by the simplices matrix faces.

    If faces is None, then the parameters must arrive in the form of a flattened (n x 3 x 2) matrix
    where n is the number of triangles. Otherwise, the faces matrix must be either (n x 3) or (n x 3
    x s); if the former, each row must list the vertex indices for the faces where the vertex matrix
    is presumed to be shaped (V x 2). Alternately, faces may be a full (n x 3 x 2) simplex array of
    the indices into the parameters.

    The optional argument axis (default: 1) may be set to 0 if the faces argument is a matrix but
    the coordinate matrix will be (2 x V) instead of (V x 2).
    '''
    faces = np.asarray(faces)
    if len(faces.shape) == 2:
        if faces.shape[1] != 3: faces = faces.T
        n = 2 * (np.max(faces) + 1)
        if axis == 0: tmp = np.reshape(np.arange(n), (2,-1)).T
        else:         tmp = np.reshape(np.arange(n), (-1,2))
        faces = np.reshape(tmp[faces.flat], (-1,3,2))
    faces = faces.flatten()
    return compose(TriangleSignedArea2DPotential(), part(Ellipsis, faces))
@pimms.immutable
class TriangleArea2DPotential(PotentialFunction):
    '''
    TriangleArea2DPotential(n) yields a potential function that tracks the unsigned area of the
    given number of faces.
    '''
    def value(self, p):
        # transpose to be 3 x 2 x n
        p = np.transpose(np.reshape(p, (-1, 3, 2)), (1,2,0))
        # First, get the two legs...
        (dx_ab, dy_ab) = p[1] - p[0]
        (dx_ac, dy_ac) = p[2] - p[0]
        (dx_bc, dy_bc) = p[2] - p[1]
        # now, the area is half the z-value of the cross-product...
        sarea0 = 0.5 * (dx_ab*dy_ac - dx_ac*dy_ab)
        # but we want to abs it
        return np.abs(sarea0)
    def jacobian(self, p, into=None):
        # transpose to be 3 x 2 x n
        p = np.transpose(np.reshape(p, (-1, 3, 2)), (1,2,0))
        # First, get the two legs...
        (dx_ab, dy_ab) = p[1] - p[0]
        (dx_ac, dy_ac) = p[2] - p[0]
        (dx_bc, dy_bc) = p[2] - p[1]
        # now, the area is half the z-value of the cross-product...
        sarea0 = 0.5 * (dx_ab*dy_ac - dx_ac*dy_ab)
        # but we want to abs it
        dsarea0 = np.sign(sarea0)
        z = np.transpose([[-dy_bc,dx_bc], [dy_ac,-dx_ac], [-dy_ab,dx_ab]], (2,0,1))
        z = times(0.5*dsarea0, z)
        m = numel(p)
        n = p.shape[2]
        ii = (np.arange(n) * np.ones([6, n])).T.flatten()
        z = sps.csr_matrix((z.flatten(), (ii, np.arange(len(ii)))), shape=(n, m))
        return safe_into(into, z)
def face_areas(faces, axis=1):
    '''
    face_areas(faces) yields a potential function f(x) that calculates the unsigned area of each
      faces represented by the simplices matrix faces.

    If faces is None, then the parameters must arrive in the form of a flattened (n x 3 x 2) matrix
    where n is the number of triangles. Otherwise, the faces matrix must be either (n x 3) or (n x 3
    x s); if the former, each row must list the vertex indices for the faces where the vertex matrix
    is presumed to be shaped (V x 2). Alternately, faces may be a full (n x 3 x 2) simplex array of
    the indices into the parameters.

    The optional argument axis (default: 1) may be set to 0 if the faces argument is a matrix but
    the coordinate matrix will be (2 x V) instead of (V x 2).
    '''
    faces = np.asarray(faces)
    if len(faces.shape) == 2:
        if faces.shape[1] != 3: faces = faces.T
        n = 2 * (np.max(faces) + 1)
        if axis == 0: tmp = np.reshape(np.arange(n), (2,-1)).T
        else:         tmp = np.reshape(np.arange(n), (-1,2))
        faces = np.reshape(tmp[faces.flat], (-1,3,2))
    faces = faces.flatten()
    return compose(TriangleArea2DPotential(), part(Ellipsis, faces))



####################################################################################################
# neuropythy/mri/core.py
# Simple tools for dealing with the cortical objects and general MRImage data
# By Noah C. Benson

import numpy               as np
import numpy.linalg        as npla
import scipy               as sp
import scipy.sparse        as sps
import scipy.spatial       as spspace
import pyrsistent          as pyr
import collections         as colls
import os, sys, types, six, pimms

from .. import geometry as geo

from itertools import chain

from ..util import (ObjectWithMetaData, to_affine, apply_affine, is_image, is_address, is_tuple,
                    is_list, address_data, address_interpolate, curry, to_hemi_str, is_pseudo_path,
                    pseudo_path, to_pseudo_path)
from .images import (to_image, to_image_spec, is_image_spec, is_image_header, is_pimage, is_npimage,
                     image_copy)

@pimms.immutable
class Subject(ObjectWithMetaData):
    '''
    Subject is a class that tracks information about an individual subject. A Subject object keeps
    track of hemispheres (Cortex objects) as well as some information about the voxels (e.g., the
    gray_mask).

    When declaring a subject, the hemispheres argument (hemis) should always include at least the
    keys 'lh' and 'rh'.
    Images should, at a minimum, include the following in their images dictionary:
      * lh_gray_mask
      * rh_gray_mask
      * lh_white_mask
      * rh_white_mask
    Alternately, the values with the same names may be overloaded in a daughter class.

    Subject respects laziness in the hemis and images classes, and this mechanism is recommended
    as a way to lazily load subject data (see pimms.lazy_map).
    '''
    def __init__(self, name=None, pseudo_path=None, hemis=None, images=None, meta_data=None):
        self.name                   = name
        self.pseudo_path            = pseudo_path
        self.hemis                  = hemis
        self.images                 = images
        self.meta_data              = meta_data

    @pimms.param
    def name(nm):
        '''
        sub.name is the name of the subject sub.
        '''
        if nm is None or pimms.is_str(nm): return nm
        else: raise ValueError('subject names must be strings or None')
    @pimms.param
    def pseudo_path(pd):
        if pd is None: return None
        else: return to_pseudo_path(pd)
    @pimms.value
    def path(pseudo_path):
        '''
        sub.path is the path of the subject's data directory, if any.
        '''
        if pseudo_path is None: return None
        return pseudo_path.actual_source_path
    @pimms.param
    def hemis(h):
        '''
        sub.hemis is a persistent map of hemisphere names ('lh', 'rh', possibly others) for the
        given subject sub.
        '''
        if   h is None:        return pyr.m()
        elif pimms.is_pmap(h): return h
        elif pimms.is_map(h):  return pyr.pmap(h)
        else: raise ValueError('hemis must be a mapping')
    @pimms.param
    def images(imgs):
        '''
        sub.images is a persistent map of MRImages tracked by the given subject sub.
        '''
        if   imgs is None:        return pyr.m()
        elif pimms.is_pmap(imgs): return imgs
        elif pimms.is_map(imgs):  return pyr.pmap(imgs)
        else: raise ValueError('images must be a mapping')

    # Updater operations:
    def with_hemi(self, *ms, **kw):
        '''
        sub.with_hemi(name=hemi) adds the give Cortex object hemi to the subject's hemis map with
          the given name.
        sub.with_hemi(name1=hemi1, name2=hemi2...) adds all the given hemispheres.
        sub.with_hemi({name1: hemi1, name2:hemi2}) is equivalent to the previous line.
        
        Note that any number of maps may be passed followed by any number of keyword arguments.
        '''
        hh = pimms.merge(self.hemis, *(ms + (kw,)))
        if hh is self.hemis: return self
        return self.copy(hemis=hh)
    def with_image(self, *ms, **kw):
        '''
        sub.with_image(name=img) adds the give Cortex object hemi to the subject's hemis map with
          the given name.
        sub.with_image(name1=img1, name2=img2...) adds all the given hemispheres.
        sub.with_image({name1:img1, name2:img2}) is equivalent to the previous line.
        
        Note that any number of maps may be passed followed by any number of keyword arguments.
        '''
        ims = pimms.merge(self.images, *(ms + (kw,)))
        if ims is self.images: return self
        return self.copy(images=ims)
    def wout_hemi(self, *args):
        '''
        sub.wout_hemi(heminame) yields a clone of the given subject object sub but with the 
          hemisphere with the given name removed.

        Note that any number of names may be given.
        '''
        hh = self.hemis
        for a in args: hh = hh.discard(a)
        return self if hh is self.hemis else self.copy(hemis=hh)
    def wout_image(self, *args):
        '''
        sub.wout_image(imgname) yields a clone of the given subject object sub but with the image
          with the given name removed.

        Note that any number of names may be given.
        '''
        ims = self.images
        for a in args: ims = ims.discard(a)
        return self if ims is self.images else self.copy(images=ims)


    # Aliases for hemispheres
    @pimms.value
    def LH(hemis):
        '''
        sub.LH is an alias for sub.hemis['lh'].
        '''
        return hemis.get('lh', None)
    @pimms.value
    def RH(hemis):
        '''
        sub.RH is an alias for sub.hemis['rh'].
        '''
        return hemis.get('rh', None)
    @pimms.value
    def lh(hemis):
        '''
        sub.lh is an alias for sub.hemis['lh'].
        '''
        return hemis.get('lh', None)
    @pimms.value
    def rh(hemis):
        '''
        sub.rh is an alias for sub.hemis['rh'].
        '''
        return hemis.get('rh', None)
    def to_hemi(self, h):
        '''
        sub.to_hemi(arg) attempts to convert arg into a hemisphere of the given subject. If arg is
          a tuple or list of hemisphere objects or names, then each is converted and returned as
          a tuple.

        Note that the objects Ellipsis and None are converted into ('lh', 'rh'). If h is a string
        that refers to a suffix of a left and right hemisphere in sub, then (lh, rh) are returned
        for that particular hemisphere (e.g., with HCP subjects, 'LR32k' will return the 32k left
        and right hemispheres as a tuple).
        '''
        if isinstance(h, Cortex): return h
        elif pimms.is_str(h):
            if h in self.hemis: return self.hemis[h]
            elif h == 'lr' or len(h) == 0: return self.to_hemi(('lh', 'rh'))
            elif h.startswith('lr_'): return self.to_hemi(('lh'+h[2:], 'rh'+h[2:]))
            # see if there is a suffix for lh and rh:
            if ('lh%s' % h) in self.hemis and ('rh%s' % h) in self.hemis:
                return (self.hemis['lh%s' % h], self.hemis['rh%s' % h])
            elif h[0] != '_' and ('lh_%s' % h) in self.hemis and ('rh_%s' % h) in self.hemis:
                return (self.hemis['lh_%s' % h], self.hemis['rh_%s' % h])
            # try converting to a hemi string
            hh = to_hemi_str(h)
            if h == hh: raise ValueError('Could not convert "%s" into a hemisphere' % h)
            else: return self.to_hemi(hh)
        elif pimms.is_vector(h):
            return tuple([self.to_hemi(u) for u in h])
        else:
            # try to convert to a hemi-string
            return self.to_hemi(to_hemi_str(h))
        raise ValueError('Unrecognized hemisphere argument: %s' % (h,))

    # Aliases for images
    @pimms.value
    def lh_gray_mask(images):
        '''
        sub.lh_gray_mask is an alias for sub.images['lh_gray_mask'].
        '''
        return images.get('lh_gray_mask', None)
    @pimms.value
    def rh_gray_mask(images):
        '''
        sub.rh_gray_mask is an alias for sub.images['rh_gray_mask'].
        '''
        return images.get('rh_gray_mask', None)
    @pimms.value
    def lh_white_mask(images):
        '''
        sub.lh_white_mask is an alias for sub.images['lh_white_mask'].
        '''
        return images.get('lh_white_mask', None)
    @pimms.value
    def rh_white_mask(images):
        '''
        sub.rh_white_mask is an alias for sub.images['rh_white_mask'].
        '''
        return images.get('rh_white_mask', None)

    @pimms.value
    def lh_gray_indices(lh_gray_mask):
        '''
        sub.lh_gray_indices is equivalent to numpy.where(sub.lh_gray_mask).
        '''
        if lh_gray_mask is None: return None
        if is_image(lh_gray_mask): lh_gray_mask = lh_gray_mask.dataobj
        return tuple([pimms.imm_array(x) for x in np.where(lh_gray_mask)])
    @pimms.value
    def rh_gray_indices(rh_gray_mask):
        '''
        sub.rh_gray_indices is equivalent to numpy.where(sub.rh_gray_mask).
        '''
        if rh_gray_mask is None: return None
        if is_image(rh_gray_mask): rh_gray_mask = rh_gray_mask.dataobj
        return tuple([pimms.imm_array(x) for x in np.where(rh_gray_mask)])
    @pimms.value
    def gray_indices(lh_gray_indices, rh_gray_indices):
        '''
        sub.gray_indices is equivalent to numpy.where(sub.gray_mask).
        '''
        if lh_gray_indices is None and rh_gray_indices is None: return None
        elif lh_gray_indices is None: return rh_gray_indices
        elif rh_gray_indices is None: return lh_gray_indices
        else:
            gis = tuple([np.concatenate(pair) for pair in zip(lh_gray_indices, rh_gray_indices)])
            for row in gis: row.setflags(write=False)
            return gis
    @pimms.value
    def lh_white_indices(lh_white_mask):
        '''
        sub.lh_white_indices is a frozenset of the indices of the white voxels in the given
        subject's lh, represented as 3-tuples.
        '''
        if lh_white_mask is None: return None
        if is_image(lh_white_mask): lh_white_mask = lh_white_mask.dataobj
        idcs = np.transpose(np.where(lh_white_mask))
        return frozenset([tuple(row) for row in idcs])
    @pimms.value
    def rh_white_indices(rh_white_mask):
        '''
        sub.rh_white_indices is a frozenset of the indices of the white voxels in the given
        subject's rh, represented as 3-tuples.
        '''
        if rh_white_mask is None: return None
        if is_image(rh_white_mask): rh_white_mask = rh_white_mask.dataobj
        idcs = np.transpose(np.where(rh_white_mask))
        return frozenset([tuple(row) for row in idcs])
    @pimms.value
    def white_indices(lh_white_indices, rh_white_indices):
        '''
        sub.white_indices is a frozenset of the indices of the white voxels in the given subject
        represented as 3-tuples.
        '''
        if   lh_white_indices is None and rh_white_indices is None: return None
        elif lh_white_indices is None: return rh_white_indices
        elif rh_white_indices is None: return lh_white_indices
        else: return frozenset(lh_white_indices | rh_white_indices)
    @pimms.value
    def image_dimensions(images):
        '''
        sub.image_dimensions is a tuple of the default size of an anatomical image for the given
        subject.
        '''
        if images is None or len(images) == 0: return None
        if pimms.is_lazy_map(images):
            # look for an image that isn't lazy...
            key = next((k for k in images.iterkeys() if not images.is_lazy(k)), None)
            if key is None: key = next(images.iterkeys(), None)
        else:
            key = next(images.iterkeys(), None)
        img = images[key]
        if img is None: return None
        if is_image(img): img = img.dataobj
        return np.asarray(img).shape
    @pimms.value
    def repr(name, path):
        '''
        sub.repr is the representation string returned by sub.__repr__().
        '''
        if name is None and path is None:
            return 'Subject(<?>)'
        elif name is None:
            return 'Subject(<\'%s\'>)' % path
        elif path is None:
            return 'Subject(<%s>)' % name
        else:
            return 'Subject(<%s>, <\'%s\'>)' % (name, path)

    def __repr__(self):
        return self.repr
    def path_join(self, *args):
        '''
        sub.path_join(args...) is equivalent to sub.pseudo_path.join(sub.path, args...)
        '''
        return self.pseudo_path.join(self.path, *args)
    def load(self, filename, *args, **kw):
        '''
        sub.load(filename) attempts to load the given filename from the pseudo-path / directory
          represented by the given subject sub. Note that additional arguments and keyword arguments
          are passed verbatim to the neuropythy.load() function.

        Note that the given filename must be a relative path. It may, however, be a list or tuple of
        diretory names (as would be passed to os.path.join).
        '''
        from neuropythy.io import load
        if self.pseudo_path is None:
            raise ValueError('cannot load from subject without a pseudo_path')
        if not pimms.is_str(filename): filename = self.pseudo_path.join(*filename)
        flnm = self.pseudo_path.local_path(filename)
        return load(flnm, *args, **kw)
    def cortex_to_image(self, data, im,
                        hemi=None, method=None, fill=0, affine=Ellipsis, address=None,
                        # below are the property() args:
                        dtype=Ellipsis, outliers=None,    data_range=None,    clipped=np.inf,
                        weights=None,   weight_min=0,     null=np.nan,        transform=None,
                        mask=None,      valid_range=None, weight_transform=Ellipsis):
        '''
        sub.cortex_to_image(data, im) yields an MRImage object with the same image-spec as im but
          with the given data projected into the image. The argument im may be anything that can be
          converted to an image-spec using the to_image_spec() function. If im is an image whose
          dataobj member is not persistent and is not an array-proxy then the image itself will be
          written into.
        sub.cortex_to_image(data, im, hemi) projects the given cortical-surface data to the given
          subject's gray-matter voxels of the given hemisphere and returns the image object.
        sub.cortex_to_image((lh_data, rh_data), im) projects into both hemispheres and yields the
          object.

        Note that if no hemisphere is given and the input argument data is not a tuple like
        (lh_data, rh_data), then it must be a property name that is shared by both hemispheres. Even
        if both hemispheres have the same vertex_count, you cannot pass one vector unless a
        hemisphere name is given. By default the hemispheres used are 'lh' and 'rh'.
    
        The data argument may be a 2-tuple, in which case it is always interpreted as (lh-data,
        rh-data); the individual hemisphere data can be a vector of values or a property name. The
        data are always passed through the cortex.property() method and any option that can be given
        to property() can be given here and will be forwarded along. The following additional
        options may also be given:
          * method (default: None) specifies that a particular method should be used; valid options
            are 'linear' and 'nearest'. The 'linear' method uses linear interpolation within the
            prismatic columns of the cortex (see cortex.to_image() for more info); the nearest
            method uses the vertex on the cortex to which the given voxel is closest (using the
            lines that the vertices make through cortex). The default, None, specifies that linear
            should be used for all real/complex (inexact) data and nearest should be used for all
            other data (integers/strings/etc).
          * fill (default: 0) specifies the value to be assigned to all voxels not in the gray mask
            or voxels in the gray-mask that are missed by the interpolation method.
          * affine (default: Ellipsis) specifies the affine transformation that should be used to
            align the cortical surfaces with the voxels. If Ellipsis, then the affine transforms
            saved in the cortex objects. Note that this option overwrites these transforms if found.
          * address (default: None) may specify pre-calculated addresses for use in projecting the
            hemispheres into the image.
        '''
        # get our image:
        if pimms.is_nparray(im) and len(im.shape) >= 3: im = ny.to_image(im)
        if is_image(im) and (is_pimage(im) or not is_npimage(im)): im = image_copy(im)
        else: im = to_image(to_image_spec(im), fill=fill)
        # what hemisphere(s)?
        hemi = self.to_hemi(hemi)
        if pimms.is_vector(hemi) and len(hemi) == 1: hemi = hemi[0]
        if not pimms.is_vector(hemi):
            if   hemi.chirality == 'lh': hemi = (hemi, None)
            elif hemi.chirality == 'rh': hemi = (None, hemi)
            else:                        hemi = (hemi, hemi)
        elif len(hemi) != 2:
            raise ValueError('Exactly two hemispheres must be given')
        elif len(np.unique([None if h is None else h.chirality for h in hemi])) != 2:
            raise ValueError('One or two different chiralities must be given in hemi argument')
        # make sure they are in the right order:
        if   hemi[1] is not None and hemi[1].chirality == 'lh': hemi = (hemi[1], hemi[0])
        elif hemi[0] is not None and hemi[0].chirality == 'rh': hemi = (hemi[1], hemi[0])
        nhems = len([x for x in hemi if x])
        tr = None
        # Make the data match this format...
        if pimms.is_str(data): data = [None if h is None else data for h in hemi]
        elif nhems == 1 and len(data) == 1:
            if len(data[0]) != (hemi[0] if hemi[1] is None else hemi[1]).vertex_count:
                raise ValueError('hemi/property size mismatch')
            data = [None if h is None else data[0] for h in hemi]
        elif len(data) == 2:
            if not (data[0] is None and hemi[0] is None or len(data[0]) == hemi[0].vertex_count):
                raise ValueError('lh hemi/property size mismatch')
            if not (data[1] is None and hemi[1] is None or len(data[1]) == hemi[1].vertex_count):
                raise ValueError('rh hemi/property size mismatch')
            data = np.asarray(data)
        else:
            data = np.asarray(data)
            sh = data.shape
            if nhems == 1: data = tuple([None if k is None else data for k in hemi])
            else: raise ValueError('cannot given single array of data for multiple hemispheres')
        # gather up the property() keywords for the next step:
        kw = dict(affine=affine,         method=method,
                  dtype=dtype,           null=null,          outliers=outliers,
                  data_range=data_range, clipped=clipped,    weights=weights,
                  weight_min=weight_min, mask=mask,          valid_range=valid_range,
                  transform=transform,   weight_transform=weight_transform)
        if address is None: address = (None, None)
        elif not pimms.is_vector(address):
            if nhems == 1: address = [None if h is None else address for h in hemi]
            else: raise ValueError('Two hemispheres but only 1 address given')
        # okay, we just need to pass down to the to_image function of the hemispheres:
        for (dat,h,addr,ii) in zip(data, hemi, address, np.arange(nhems)):
            if h is None: dat = None
            if dat is None: continue
            if addr is not None: kw['address'] = addr
            elif 'address' in kw: del kw['address']
            im = h.to_image(dat, im, **kw)
        return im
    def image_to_cortex(self, image,
                        surface='midgray', hemi=None, affine=Ellipsis, method=None, fill=0,
                        dtype=None, weights=None):
        '''
        sub.image_to_cortex(image) is equivalent to the tuple
          (sub.lh.from_image(image), sub.rh.from_image(image)).
        sub.image_to_cortex(image, surface) uses the given surface (see also cortex.surface).
        '''
        if hemi is None: hemi = 'both'
        hemi = hemi.lower()
        if hemi in ['both', 'lr', 'all', 'auto']:
            return tuple(
                [self.image_to_cortex(image, surface=surface, hemi=h, affine=affine,
                                      method=method, fill=fill, dtype=dtype, weights=weights)
                 for h in ['lh', 'rh']])
        else:
            hemi = getattr(self, hemi)
            return hemi.from_image(image, surface=surface, affine=affine,
                                   method=method, fill=fill, dtype=dtype, weights=weights)
    def image_address(self, image, hemi='lr'):
        '''
        sub.image_address(image) yields the (lh, rh) image-addresses for the given image or
          image-spec.
        sub.image_address(image, h) uses the given hemisphere h, which may alternately be a tuple of
          hemisphere names. The default, 'lr', is equivalent to ('lh','rh').
        '''
        if is_tuple(hemi):
            sing = False
            hemi = [to_hemi_str(h) for h in hemi]
            if h == ['lr']: h = ['lh','rh']
        else:
            sing = True
            hemi = to_hemi_str(hemi)
            hemi = ['lh','rh'] if hemi == 'lr' else [hemi]
        r = [self.hemis[h].image_address(image)
             for h in hemi]
        if sing and len(r) == 1: return r[0]
        else: return tuple(r)
def is_subject(s):
    '''
    is_subject(s) yields True if s is a Subject object and False otherwise.
    '''
    return isinstance(s, Subject)

@pimms.immutable
class Cortex(geo.Topology):
    '''
    Cortex is a class that handles a single cortical hemisphere; cortex tracks both the spherical
    registrations that can be used for cross-subject interpolation as well as the various surfaces
    and cortical layers that can be produced from the combined white/pial surfaces.
    Cortex is the go-to class for performing interpolation between subjects as it is a Topology
    object and thus knows how to search for common registrations for interpolation and comparison.
    Cortex also holds the required methods for creating map projections from spherical
    registrations.

    Cortex(chirality, tess, surfaces, registrations) is typically used to initialize a cortex
    object. The chirality should be either 'lh' or 'rh'; tess must be the tesselation of the cortex
    object. The surfaces and registrations arguments should both be (possibly lazy) maps whose
    values are the appropriate mesh objects. The surfaces must include 'white' and 'pial'. If the
    registrations includes the key 'native' this is taken to be the default registration for the
    particular cortex object.
    '''
    def __init__(self, chirality, tess, surfaces, registrations,
                 properties=None, affine=None, meta_data=None):
        self.chirality = chirality
        self.surface_coordinates = surfaces
        self.meta_data = meta_data
        geo.Topology.__init__(self, tess, registrations, properties=properties)
        self.chirality = chirality
        self.surface_coordinates = surfaces
        self.affine = affine
        self.meta_data = meta_data
    @pimms.param
    def chirality(ch):
        '''
        cortex.chirality gives the chirality ('lh' or 'rh') for the given cortex.
        '''
        if ch is None: return None
        ch = ch.lower()
        if ch != 'lh' and ch != 'rh':
            raise ValueError('chirality must be \'lh\' or \'rh\'')
        return ch
    @pimms.param
    def affine(aff):
        '''
        cortex.affine is either None or an affine transformation that specifies how the coordinates
        of the surfaces in the given cortex should be rotated in order to align with an abstract
        'native' geometry. This field is used, e.g., by FreeSurfer subjects to indicate how the
        coordinates of the FreeSurfer surfaces should be transformed in order to be aligned with the
        native-space defined by the affine transforms of the FreeSurfer images.
        '''
        if aff is None: return None
        return to_affine(aff, 3)
    @pimms.param
    def surface_coordinates(surfs):
        '''
        cortex.surface_coordinates is a mapping of the surface coordinates of the given cortex; this
        must include the surfaces 'white' and 'pial'.
        '''
        if pimms.is_map(surfs):
            return pimms.persist(surfs)
        else:
            raise ValueError('surface_coordinates must be a mapping object')
    @pimms.value
    def surfaces(surface_coordinates, properties, tess, chirality):
        '''
        cortex.surfaces is a mapping of the surfaces of the given cortex; this must include the
        surfaces 'white' and 'pial'.
        '''
        def _make_mesh(name):
            def _lambda():
                val = surface_coordinates[name]
                if isinstance(val, geo.Mesh): val = val.coordinates
                m = geo.Mesh(tess, val, properties=properties, meta_data={'chirality':chirality})
                return m.persist()
            return _lambda
        return pimms.lazy_map({k:_make_mesh(k) for k in six.iterkeys(surface_coordinates)})
    @pimms.value
    def aligned_surfaces(surfaces, affine):
        '''
        cortex.aligned_surfaces is identical to cortex.surfaces except that the aligned surfaces
        have been transformed by cortex.affine.
        '''
        if affine is None: return surfaces
        def align_surf(s):
            srf = surfaces[s]
            return srf.copy(coordinates=apply_affine(affine, srf.coordinates))
        return pimms.lazy_map({k:curry(align_surf, k) for k in six.iterkeys(surfaces)})
    @pimms.require
    def validate_surfaces(surfaces):
        '''
        validate_surfaces requires that the surfaces map contain the keys 'white' and 'pial'.
        '''
        if 'white' in surfaces and 'pial' in surfaces: return True
        else: raise ValueError('surfaces parameter must contain both \'white\' and \'pial\'')

    @pimms.value
    def white_surface(surfaces):
        '''
        cortex.white_surface is the mesh representing the white-matter surface of the given cortex.
        '''
        return surfaces['white']
    @pimms.value
    def pial_surface(surfaces):
        '''
        cortex.pial_surface is the mesh representing the pial surface of the given cortex.
        '''
        return surfaces['pial']
    @pimms.value
    def midgray_surface(white_surface, pial_surface):
        '''
        cortex.midgray_surface is the mesh representing the midgray surface half way between the
        white and pial surfaces.
        '''
        midgray_coords = 0.5*(white_surface.coordinates + pial_surface.coordinates)
        return white_surface.copy(coordinates=midgray_coords)
    @pimms.value
    def white_to_pial_vectors(white_surface, pial_surface):
        '''
        cortex.white_to_pial_vectors is a (3 x n) matrix of the unit direction vectors that point
          from the n vertices in the cortex's white surface to their equivalent positions in the
          pial surface.
        '''
        u = pial_surface.coordinates - white_surface.coordinates
        d = np.sqrt(np.sum(u**2, axis=0))
        z = np.isclose(d, 0)
        return pimms.imm_array(np.logical_not(z) / (d + z))
    @pimms.value
    def repr(chirality, tess, vertex_count):
        '''
        cortex.repr is equivalent to repr(cortex).
        '''
        arg = ('XH' if chirality is None else chirality.upper(), tess.face_count, vertex_count)
        return 'Cortex(<%s>, <%d faces>, <%d vertices>)' % arg

    def __repr__(self):
        return self.repr
    def surface(self, name='white', aligned=False):
        '''
        cortex.surface() yields the white surface of the given cortex
        cortex.surface(name) yields the surface with the given name (e.g., 'white', 'pial',
          'inflated', 'midgray').
        cortex.surface(fraction) yields the surface that is <fraction> of the distance from white
          to pial (0 is equivalent to 'white'; 1 is equivalent to 'pial'). Layers outside of the
          range 0-1 may be returned by following the vectors between white and pial surfaces, but
          they may have odd appearances, and this should not be confused with surface inflation.
        cortex.surface([dist]) yields the layer that is the given distance from the white surface.

        The optional argument aligned (default: False) may be specified to extract meshes already
        aligned according to the cortex.affine matrix.
        '''
        srfs = self.aligned_surfaces if aligned else self.surfaces
        if pimms.is_str(name):
            if name.lower() == 'midgray' and 'midgray' not in srfs:
                return self.midgray_surface # in case it's been provided via overloading
            elif name in srfs: return srfs[name]
            else:
                m = geo.to_mesh((self, name))
                if not aligned or self.affine is None: return m
                return m.copy(coordinates=apply_affine(self.affine, m.coordinates))
        elif pimms.is_vector(name, 'real') and len(name) == 1:
            x0 = srfs['white'].coordinates
            x1 = srfs['pial'].coordinates
            q = name[0]
            return self.make_mesh(x0*(1 - q) + x1*q)
        elif pimms.is_real(name):
            x0 = srfs['white'].coordinates
            x1 = srfs['pial'].coordinates
            return self.make_mesh((1 - name)*x0 + name*x1)
        else: raise ValueError('could not understand surface layer: %s' % name)
    def from_image(self, image,
                   surface='midgray', affine=Ellipsis, method=None, fill=0, dtype=None,
                   weights=None):
        '''
        cortex.from_image(image) is equivalent to cortex.midgray_surface.from_image(image).
        cortex.from_image(image, surface) uses the given surface (see also cortex.surface).

        The optional argument affine (default: Ellipsis) can be used to override the cortex.affine
        transformation typically used in cortex-image alignment.
        '''
        mesh = geo.to_mesh((self, surface))
        if affine is Ellipsis: affine = self.affine
        return mesh.from_image(image, affine=affine, method=method, fill=fill, dtype=dtype,
                               weights=weights)
    def to_image(self, prop, image,
                 fill=0, image_type=None, method=None, address=None, affine=Ellipsis,
                 # the rest of the arguments are for property()
                 dtype=Ellipsis,   outliers=None,  data_range=None,           clipped=np.inf,
                 weights=None,     weight_min=0,   weight_transform=Ellipsis, mask=None,
                 valid_range=None, null=np.nan,    transform=None):
        '''
        cortex.to_image(property, ...) yields a 3D image of the given property.
        cortex.to_image(property, image) writes the data into the given image and returns it if the
          image contains a writeable numpy array; otherwise duplicates the image with a new
          read-only array that consists of the new data written over the old data and yields that
          image. Note that most images returned by nibabel use array proxies and thus are duplicated
          rather than written to by default (because the dataobj is an ArrayProxy and not a
          writeable numpy array).
        cortex.to_image(property, imspec) uses the given image-specification to construct the image.

        Additional keyword arguments accepted by cortex.property() are passed along when determining
        the property vector. The following additional optional arguments are also accepted:
          * fill (default: 0) also specifies the value given to the cells of a new image if one must
            be constructed (for example, if no image is passed or an image-spec is given).
          * image_type (default: None) specifies the image type that should be returned; by default
            uses the type of the given image or 'nifti1' if no image type can be inferred.
          * method (default: None) specifies the method of interpolation to use. By default, this
            is 'linear' if the property data are inexact numbers and 'nearest' if the data are
            anything else (strings, integers, etc.).
          * address (default: None) may optionally be used to provide the pre-calculated addresses
            rather than using cortex.image_address() to find them.
        '''
        # start by getting the propery:
        propkw = dict(dtype=dtype,           null=null,                         outliers=outliers,
                      data_range=data_range, clipped=clipped,                   weights=weights,
                      mask=mask,             valid_range=valid_range,           transform=transform,
                      weight_min=weight_min, weight_transform=weight_transform)
        if pimms.is_map(prop):
            ktr = lambda k:0 if k == 'white' else 1 if k == 'pial' else 0.5 if k == 'midgray' else k
            prop = {ktr(k):self.property(v, **propkw) for (k,v) in six.iteritems(prop)}
        else:
            prop = self.property(prop, **propkw)
            # we assume these are linearly-spaced between 0 and 1
            if pimms.is_vector(prop): prop = {0:prop, 1:prop}
            else: prop = {k:v for (k,v) in zip(np.linspace(0, 1, len(prop)), prop)}
        # next, figure out the image
        if   is_image_spec(image):   image = to_image(image,
                                                      fill=fill, image_type=image_type, dtype=dtype)
        elif is_image_header(image): image = to_image(image,
                                                      fill=fill, image_type=image_type, dtype=dtype)
        else:                        image = to_image(image, image_type=image_type, dtype=dtype)
        if is_pimage(image) or not pimms.is_nparray(image.dataobj):
            image = image_copy(image, image_type=image_type)
        if image_type is not None:
            image_type = to_image_type(image_type)
            if to_image_type(image) is not image_type:
                image = image_copy(image, image_type=image_type)
        # okay, now we get the projection of the cortex into the image:
        addr = self.image_address(image, affine=affine) if address is None else address
        # and use this to interpolate...
        dat = address_interpolate(addr, prop, method=method, null=null)
        # and put these into the image!
        image.dataobj[tuple(addr['voxel_indices'])] = dat
        return image
    def image_address(self, img, affine=Ellipsis):
        '''
        cortex.image_address(img) is like cortex.address() but works on images; note that
          cortex.address(img) will call cortex.address_image(img) when img is an image.

        The optional argument affine may be used to override the cortex.affine transform usually
        used in cortex-image alignment.

        Note that the algorithm employed by this method works by dividing each triangular "prism"
        (colum through the cortex defined by each face in the cortex tesselation) into a number of
        tetrahedrons; because this definition of cortex is not necessarily identical to other 
        definitions (such as that used by FreeSurfer itself), so depending the precise definition
        of what is inside the cortex in a particular image may vary around the edges.
        '''
        dat = to_image_spec(img)
        aff = dat['affine']
        imsh = np.reshape(dat['image_shape'], (3,1))
        if affine is Ellipsis: affine = self.affine
        if affine is not None: aff = np.dot(np.linalg.inv(affine), aff)
        aff = np.linalg.inv(aff)
        # transform coordinates into voxel-space:
        (wx,px) = (self.white_surface.coordinates, self.pial_surface.coordinates)
        (wx,px) = [np.dot(aff, np.vstack([x, np.ones([1,x.shape[1]])]))[:3] for x in (wx,px)]
        # now we find the addresses... start by getting the face coords:
        fs = self.tess.indexed_faces
        (wfx,pfx) = [np.transpose(x[:,fs], (1,0,2)) for x in (wx,px)]
        # okay, make a bounding box for each prism in the cortex:
        x = np.vstack([wfx,pfx])
        (mn,mx) = [f(x, axis=0) for f in (np.min,np.max)]
        mn[mn < 0] = 0
        for (ii,n) in enumerate(imsh): mx[ii,mx[ii] >= n] = n - 1
        # number of voxels in each bounding box:
        (mn, mx) = (np.ceil(mn), np.floor(mx) + 1)
        dims = mx - mn
        dims[dims < 0] = 0
        nvox = np.prod(dims, axis=0).astype('int') #dbg
        ii = np.where(nvox > 0)[0]
        # now we build up a big list of voxel indices and prisms to test against each other:
        i1 = np.cumsum(nvox[ii])
        i0 = np.concatenate([[0], i1[:-1]])
        n  = int(i1[-1])
        idcs = np.zeros((3,n),   dtype=np.int)
        wfxs = np.zeros((3,3,n), dtype=np.float)
        pfxs = np.zeros((3,3,n), dtype=np.float)
        iis  = np.zeros(n,       dtype=np.int)
        # we step along from i0 to i1 forgetting finished prisms along the way
        (kk, q, mn, nvox, dims) = (ii, 0, mn[:,ii], nvox[ii], dims[:,ii])
        while len(kk) > 0:
            wfxs[:,:,i0] = wfx[:,:,kk]
            pfxs[:,:,i0] = pfx[:,:,kk]
            iis[i0] = kk
            idcs[:,i0] = [(q // (dims[1]*dims[2]))      + mn[0],
                          np.mod(q // dims[2], dims[1]) + mn[1],
                          np.mod(q, dims[2])            + mn[2]]
            # next iteration:
            q += 1
            ki = np.where(nvox > q)[0]
            nvox = nvox[ki]
            kk = kk[ki]
            if len(kk) == 0: break
            mn = mn[:,ki]
            dims = dims[:,ki]
            i0 = i0[ki]
            i0 += 1
        # now test them all:
        bcs = geo.prism_barycentric_coordinates(wfxs, pfxs, idcs)
        ok = ~np.isclose(np.sum(bcs, axis=0), 0)
        # potentially-useful debug code (not currently used)
        #object.__setattr__(self, '_debug', {}) #dbg
        #self._debug['mn'] = np.array(mn) #dbg
        #self._debug['mx'] = np.array(mx) #dbg
        #self._debug['dims'] = np.array(dims) #dbg
        #self._debug['i0'] = i0 #dbg
        #self._debug['bcs'] = bcs
        #self._debug['ok'] = ok
        #self._debug['wfxs'] = wfxs
        #self._debug['pfxs'] = pfxs
        #self._debug['idcs'] = idcs
        idcs = idcs[:,ok]
        bcs = bcs[:,ok]
        ii = iis[ok]
        #(ii,kk) = np.unique(ii, return_index=True)
        #return {'coordinates':bcs[:,kk], 'faces':self.tess.faces[:,ii], 'voxel_indices':idcs[:,kk]}
        return {'coordinates':bcs, 'faces':self.tess.faces[:,ii], 'voxel_indices':idcs}
    def address(self, data, affine=Ellipsis):
        '''
        cortex.address(points) yields the barycentric coordinates of the given point or points; the
          return value is a dict whose keys are 'face_id' and 'coordinates'. The address may be used
          to interpolate or unaddress either from a surface mesh or from a cortex.
        cortex.address(image) is equivalent to cortex.address(image, mask) where mask is equivalent
          to (numpy.isfinite(image) & image.astype(numpy.bool)).

        The optional argument affine (default: Ellipsis) may be set to an affine transformation that
        should be applied prior to aligning the cortex with an image (if a point-set is given, then
        this parameter is ignored). Ellipsis indicates that cortex.affine should be used.
        '''
        if is_image(data): return self.image_address(data, affine=affine)
        data = np.asarray(data) # data must be a point matrix then
        if len(data.shape) > 2: raise ValueError('point or point matrix required')
        if len(data.shape) == 2: xyz = data.T if data.shape[0] == 3 else data
        else:                    xyz = np.asarray([data])
        if not xyz.flags['WRITEABLE']: xyz = np.array(xyz)
        # now, get the barycentric coordinates...
        n = len(xyz)
        fcount = self.tess.face_count
        # get some relevant structure data
        (wsrf, psrf) = (self.white_surface, self.pial_surface)
        (fwcoords,fpcoords) = (wsrf.face_coordinates, psrf.face_coordinates)
        fids = range(fcount)
        fids = np.concatenate((fids, fids))
        faces = np.concatenate((self.tess.indexed_faces, self.tess.indexed_faces), axis=-1)
        face_centers = np.hstack((wsrf.face_centers, psrf.face_centers))
        try:              shash = spspace.cKDTree(face_centers.T)
        except Exception: shash = spspace.KDTree(face_centers.T)
        (whsh, phsh) = [s.face_hash for s in (wsrf, psrf)]
        # we can define a max distance for when something is too far from a point to plausibly be
        # in a prism:
        wpdist = np.sqrt(np.sum((wsrf.coordinates - psrf.coordinates)**2, axis=0))
        max_dist = np.max(np.concatenate([wsrf.edge_lengths, psrf.edge_lengths, wpdist]))
        # Okay, for each voxel (xyz), we want to find the closest face centers; from those
        # centers, we find the ones for which the nearest point in the plane of the face to the
        # voxel lies inside the face, and of those we find the closest; this nearest point is
        # then used for trilinear interpolation of the points in the prism.
        (N, sofar) = (256, 0)
        # go ahead and make the results
        res_fs = np.full((3, n),     -1, dtype=np.int)
        res_xs = np.full((3, n), np.nan, dtype=np.float)
        # points tha lie outside the pial surface entirely we can eliminate off the bat:
        ii = [(mn <= ix) & (ix <= mx)
              for (px,ix) in zip(psrf.coordinates, xyz.T)
              for (mn,mx) in [(np.min(px), np.max(px))]]
        ii = np.where(ii[0] & ii[1] & ii[2])[0]
        # Okay, we look for those isect's within the triangles
        idcs = []
        for i in range(N):
            if len(ii) == 0: break
            if i >= sofar:
                sofar = max(4, 2*sofar)
                (ds,idcs) = shash.query(xyz[ii], sofar)
            # if dist is greater than max distance, we can skip those
            oks = np.where(ds[:,i] <= max_dist)[0]
            if len(oks) == 0: break
            elif len(oks) < len(ii):
                ii = ii[oks]
                idcs = idcs[oks]
                ds = ds[oks]
            col = fids[idcs[:,i]]
            bcs = geo.prism_barycentric_coordinates(fwcoords[:,:,col], fpcoords[:,:,col], xyz[ii].T)
            # figure out which ones were discovered to be in this prism
            outp = np.isclose(np.sum(bcs, axis=0), 0)
            if not np.all(outp):
                inp    = np.logical_not(outp)
                ii_inp = ii[inp]
                # for those in their prisms, we capture the face id's and coordinates
                res_fs[:, ii_inp] = faces[:, col[inp]]
                res_xs[:, ii_inp] = bcs[:, inp]
                # trim down those that matched so we don't keep looking for them
                ii = ii[outp]
                idcs = idcs[outp]
                ds = ds[outp]
            # And continue!
        # and return the data
        return {'faces': res_fs, 'coordinates': res_xs}
    def unaddress(self, data, surface=0.5):
        '''
        cortex.unaddress(address) yields the (3 x n) coordinate matrix of the given addresses (or,
          if address is singular, the 3D vector) in the given cortex. If the address is a 2D instead
          of a 3D address, then the mid-gray position is returned by default.

        The following options may be given:
          * surface (default: 0.5) specifies the surface to use for 2D addresses; this should be
            either 'white', 'pial', 'midgray', or a real number in the range [0,1] where 0 is the
            white surface and 1 is the pial surface.
        '''
        (faces, coords) = address_data(data, 3, surface=surface)
        (bc, ds) = (coords[:2], coords[2])
        faces = self.tess.index(faces)
        (wx, px) = (self.white_surface.coordinates, self.pial_surface.coordinates)
        if all(len(np.shape(x)) > 1 for x in (faces, coords)):
            (wtx, ptx) = [
                np.transpose([sx[:,ff] if ff[0] >= 0 else null for ff in faces.T], (2,1,0))
                for null in [np.full((3, wx.shape[0]), np.nan)]
                for sx   in (wx, px)]
        elif faces == -1:
            return np.full(selfx.shape[0], np.nan)
        else:
            (wtx, ptx) = [sx[:,faces].T for sx in (wx, px)]
        (wu, pu) = [geo.barycentric_to_cartesian(tx, bc) for tx in (wtx, ptx)]
        return wu*ds + pu*(1 - ds)
    def image_weight(self, img, affine=None):
        '''
        cortex.image_weight(img) yields an image with the spec given by img (which may be an image
          or an image-spec) and with each voxel containing the weight of the cortex in that voxel.
          The weight is equal to the fraction of the voxel's volume that overlaps with the cortex.

        Note: not yet implemented.
        '''
        raise NotImplementedError('image_weight is not yet implemented') #TODO

def is_cortex(c):
    '''
    is_cortex(c) yields True if c is a Cortex object and False otherwise.
    '''
    return isinstance(c, Cortex)
def to_cortex(c):
    '''
    to_cortex(c) yields a Cortex object if the argument c can be coerced to one and otherwise raises
      an error.

    An object can be coerced to a Cortex object if:
      * it is a cortex object
      * it is a tuple (subject, h) where subject is a subject object and h is a subject hemisphere.
    '''
    if is_cortex(c): return c
    elif pimms.is_vector(c) and len(c) == 2:
        (s,h) = c
        if is_subject(s) and pimms.is_str(h):
            if h in s.hemis: return s.hemis[h]
            else: raise ValueError('to_cortex: hemi %s not found in given subject' % h)
    raise ValueError('Could not coerce argument to Cortex object')

####################################################################################################
# These functions deal with cortex_to_image and image_to_cortex interpolation:
def _vertex_to_voxel_linear_interpolation(hemi, gray_indices, image_shape, voxel_to_vertex_matrix):
    if gray_indices is None: raise ValueError('gray indices cannot be None')
    n      = len(gray_indices[0])
    vcount = hemi.vertex_count
    # convert voxels to vertex-space
    xyz = voxel_to_vertex_matrix.dot(np.vstack((np.asarray(gray_indices), np.ones(n))))[0:3].T
    if not xyz.flags['WRITEABLE']: xyz = np.array(xyz)
    # get some relevant structure data
    (fwcoords,fpcoords)  = (hemi.white_surface.face_coordinates, hemi.pial_surface.face_coordinates)
    fids = np.concatenate((range(hemi.tess.face_count), range(hemi.tess.face_count)))
    faces = np.concatenate((hemi.tess.indexed_faces, hemi.tess.indexed_faces), axis=-1)
    face_centers = np.hstack((hemi.white_surface.face_centers, hemi.pial_surface.face_centers))
    try:              shash = spspace.cKDTree(face_centers.T)
    except Exception: shash = spspace.KDTree(face_centers.T)
    (whsh, phsh) = [getattr(hemi, '%s_surface' % s).face_hash for s in ['white', 'pial']]
    # Okay, for each voxel (xyz), we want to find the closest face centers; from those centers, we
    # find the ones for which the nearest point in the plane of the face to the voxel lies inside
    # the face, and of those we find the closest; this nearest point is then used for trilinear
    # interpolation of the points in the triangle.
    N = 256
    sofar = 0
    # go ahead and make our interp matrix
    interp = sps.lil_matrix((n, vcount), dtype=np.float)
    # Okay, we look for those isect's within the triangles
    ii = np.asarray(range(n)) # the subset not yet matched
    for i in range(N):
        if len(ii) == 0: break
        if i >= sofar:
            sofar = max(4, 2*sofar)
            idcs = shash.query(xyz[ii], sofar)[1].T
        # we look at just the i'th column of the indices
        col = fids[idcs[i]]
        bcs = geo.prism_barycentric_coordinates(fwcoords[:,:,col], fpcoords[:,:,col], xyz[ii].T)
        bcs = bcs[0:3] # since the layers are the same in this case...
        outp = np.isclose(np.sum(bcs, axis=0), 0)
        inp = np.logical_not(outp)
        # for those in their prisms, we linearly interpolate using the bc coordinates
        bcs = bcs[:,inp]
        ii_inp = ii[inp]
        fs = faces[:, col[inp]]
        for k in (0,1,2): interp[(ii_inp, fs[k])] = bcs[k]
        # trim down those that matched so we don't keep looking for them
        ii = ii[outp]
        idcs = idcs[:,outp]
        # And continue!
    # last, we normalize the rows
    rowsums = np.asarray(interp.sum(axis=1))[:,0]
    z = np.isclose(rowsums, 0)
    invrows = np.logical_not(z) / (rowsums + z)
    ii = np.asarray(range(n))
    return sps.csc_matrix((invrows, (ii,ii))).dot(interp)
    
def _vertex_to_voxel_lines_interpolation(hemi, gray_indices, image_shape, vertex_to_voxel_matrix):
    ijks = np.asarray(list(gray_indices) if isinstance(gray_indices, colls.Set) else gray_indices)
    ijks = ijks.T if ijks.shape[0] != 3 else ijks
    n = hemi.vertex_count
    # we also need the transformation from surface to voxel
    # given that the voxels assume 0-based indexing, this means that the center of each voxel
    # is at (i,j,k) for integer (i,j,k) > 0; we want the voxels to start and end at integer
    # values with respect to the vertex positions, so we add 1/2 to the vertex
    # positions, which should make the range 0-1, for example, cover the vertices in the first
    # voxel
    tmtx = vertex_to_voxel_matrix[0:3]
    (pialX, whiteX) = [np.dot(tmtx, np.vstack((mtx, np.ones(mtx.shape[1]))))
                       for mtx in [hemi.pial_surface.coordinates, hemi.white_surface.coordinates]]
    # get vectors along each axis
    u = pialX - whiteX
    # normalize these...
    lens = np.sqrt(np.sum(u**2, axis=0))
    z = np.isclose(lens, 0)
    inv_lens = np.logical_not(z) / (lens + z)
    u = inv_lens * u
    usign = np.sign(u)
    z = usign == 0
    u_inv = np.logical_not(z) / (u + z)
    # find smallest and largest voxel indices through which a particular line passes
    (mins, maxs) = [np.array(x) for x in [whiteX, pialX]]
    # extend these a little so that we don't miss any voxels
    tmp = u*(lens*0.05)
    mins -= tmp
    maxs += tmp
    # make a lookup table of gray indices (i,j,k) to the gray voxel's index
    # this is a trick to get this all pushed into c-modules:
    ijk_mult = (image_shape[1]*image_shape[2], image_shape[2], 1)
    def ijk_to_idx(ijk):
        return ijk_mult[0]*ijk[0] + ijk_mult[1]*ijk[1] + ijk[2]
    idcs = ijk_to_idx(ijks)
    index = sps.csr_matrix(
        (range(1, 1+len(idcs)), (np.zeros(len(idcs)), idcs)),
        shape=(1,np.prod(image_shape)),
        dtype=np.int)
    # make the interpolation matrix...
    interp = sps.lil_matrix((len(ijks[0]), n), dtype=np.float)
    # ends are the voxels in which the lines end
    ends = usign * np.ceil(usign*maxs)
    # Okay, we are going to walk along each of the lines...
    ii = np.asarray(range(n))
    while len(ii) > 0:
        # we know that the current min values make a voxel on the line, but the question is, which
        # voxel will be the next along the line?
        # we want the ceiling of the current minimum; however, if the direction is negative (usign)
        # then we need to reverse it, ceiling, then reverse it again:
        z = (mins == np.round(mins))
        uu = usign * np.ceil(usign * mins) # this is, for each dim, the nearest int boundary
        # the new end-point might be farther than max? if so we replace it with max
        uu = usign * np.min((uu, usign*ends), axis=0) # usign here undoes negatives
        uu[z] += usign[z]
        d3 = (uu - mins) * u_inv # the dist in each dim from the nearest integer boundary
        # whichever row of d3 is smallest (rows are x, y, or z) determines how far the next voxel is
        min_d = np.argsort(d3, axis=0)[0]
        addrs = (min_d, range(len(min_d)))
        d = d3[addrs]
        frac = d * inv_lens
        frac[inv_lens == 0] = 1
        # what is the start voxel?
        start = np.asarray(usign * np.floor(usign * mins), dtype=np.int)
        # we want to add these fractions into the interp matrix
        tmp = ijk_to_idx(start)
        oob = np.any(start < 0, axis=0) | (tmp >= index.shape[1])
        tmp[oob] = 0
        found_idcs = index[0, tmp].toarray()[0]
        gray_fis = (found_idcs > 0) & np.logical_not(oob)
        found_idcs = found_idcs[gray_fis] - 1
        interp[(found_idcs, ii[gray_fis])] = frac[gray_fis]
        # update the min values
        mins = mins + d*u
        # if the min is now equal to the max, then we are done with that line!
        keep = np.where(np.logical_not(np.isclose(np.sum((mins - maxs)**2, axis=0), 0))
                        & np.all((maxs - mins) * u_inv > 0, axis=0))[0]
        if len(keep) < len(ii):
            (inv_lens,ii) = [x[keep] for x in (inv_lens,ii)]
            (ends,mins,maxs,usign,u,u_inv) = [x[:,keep] for x in (ends,mins,maxs,usign,u,u_inv)]
    # now we want to scale the rows by their totals
    totals = np.asarray(interp.sum(axis=1))[:,0]
    zs = np.isclose(totals, 0)
    inv_totals = np.logical_not(zs) / (totals + zs)
    rng = range(len(inv_totals))
    interp = sps.csr_matrix((inv_totals, (rng, rng))).dot(interp.tocsc())
    # That's all we have to do!
    return interp

def _vertex_to_voxel_nearest_interpolation(hemi, gray_indices, voxel_to_vertex_matrix):
    if isinstance(hemi, Subject):       hemi   = (hemi.lh, hemi.rh)
    if isinstance(hemi, geo.VertexSet): vcount = hemi.vertex_count
    else:                               vcount = hemi[0].vertex_count + hemi[1].vertex_count
    n   = len(gray_indices[0])
    xyz = voxel_to_vertex_matrix.dot(np.vstack((np.asarray(gray_indices), np.ones(n))))[0:3].T
    if not xyz.flags['WRITEABLE']: xyz = np.array(xyz)
    # find the nearest to the white and pial layers
    if isinstance(hemi, Cortex):
        (wd, wn) = hemi.white_surface.vertex_hash.query(xyz, 1)
        (pd, pn) = hemi.pial_surface.vertex_hash.query(xyz, 1)
        oth = pd < wd
        wn[oth] = pn[oth]
    else:
        (lwd, lwn) = hemi[0].white_surface.vertex_hash.query(xyz, 1)
        (lpd, lpn) = hemi[0].pial_surface.vertex_hash.query(xyz, 1)
        (rwd, rwn) = hemi[1].white_surface.vertex_hash.query(xyz, 1)
        (rpd, rpn) = hemi[1].pial_surface.vertex_hash.query(xyz, 1)
        nn = hemi[0].vertex_count
        oth = lpd < lwd
        lwn[oth] = lpn[oth]
        lwd[oth] = lpd[oth]
        oth = rpd < lwd
        lwn[oth] = rpn[oth] + nn 
        lwd[oth] = rpd[oth]
        oth = rwd < lwd
        lwn[oth] = rwn[oth] + nn
        wn = lwn
    return sps.csr_matrix((np.ones(len(wn)), (range(len(wn)), wn)),
                          shape=(len(gray_indices[0]), vcount),
                          dtype=np.float)



####################################################################################################
# neuropythy/freesurfer/__init__.py
# This file defines the FreeSurfer tools that are available as part of neuropythy.

from .core import (subject, forget_subject, forget_all, tkr_vox2ras,
                   find_subject_path, subject_paths, add_subject_path, clear_subject_paths,
                   to_mgh, load_LUT,
                   freesurfer_subject_filemap_instructions, freesurfer_subject_data_hierarchy,
                   subject_file_map, subject_dir)




####################################################################################################
# neuropythy/plans/core.py
# Core utilities for PIMMS Calculation Plan algorithms/workflows.
# By Noah C. Benson

import numpy as np
import pyrsistent as pyr
import os, sys, gzip, six, types, pimms
from ..util import (is_tuple, is_list)
from ..math import (pytorch, totensor, astensor)
from .. import math


def limit_param(param, min=-1, max=1):
    '''
    limit_param(u) yields the limited parameter x, which must always be in the
      range -1 to 1, from its unlimited value u, which may take on any real
      value.
    limit_param(u, min, max) uses the given minimum and maximum parameters in
      place of -1 and 1.

    The opposite of x = limit_param(u) is u = unlimit_param(x). The intended use
    of these functions is that, during a continuous optimization in which a
    parameter should be restricted to a particular range, the parameter over
    which the optimization occurs is in fact the unlimited parameeter, which is
    allowed to take on any real value, and which gets limited to its range
    prior to its use in any calculation.
    '''
    return min + (max - min) * (math.arctan(param)/np.pi + 0.5)
def unlimit_param(param, min=-1, max=1):
    '''
    unlimit_param(x) yields the unlimited parameter u, which may take on any
      real value, from its limited value x, which must be between -1 and 1.
    unlimit_param(x, min, max) uses the given min and max values instead of the
      default values -1 and 1.

    The opposite of u = unlimit_param(x) is x = limit_param(u). The intended use
    of these functions is that, during a continuous optimization in which a
    parameter should be restricted to a particular range, the parameter over
    which the optimization occurs is in fact the unlimited parameeter, which is
    allowed to take on any real value, and which gets limited to its range
    prior to its use in any calculation.
    '''
    return math.tan(np.pi * ((param - min) / (max - min) - 0.5))
def imap_forget(imap, ks):
    '''
    imap_forget(imap, k) yields imap after clearing the cacne for key k. This
      can also be accomplished with: del imap[k].
    imap_forget(imap, [k1, k2, ...]) clears the cache for all the given keys.
    '''
    if pimms.is_str(ks): ks = [ks]
    for k in ks:
        del imap[k]
    return imap
def imap_efferents(imap, ks):
    '''
    imap_efferents(imap, k) yields the key names of any efferent of the key k
      in imap.
    imap_efferents(imap, [k1, k2, ...]) yields the key names of the any
      efferent of any of the given keys.
    '''
    if pimms.is_str(ks): ks = [ks]
    effs = set([])
    for (k,deps) in imap.plan.dependencies.items():
        if k in ks: continue
        for d in deps:
            if d in ks:
                effs.add(k)
                break
    return list(effs)
